import ray
import logging
import json
from selector.log_setup import TournamentEncoder


@ray.remote(num_cpus=1)
class TargetAlgorithmObserver:

    def __init__(self, scenario):
        self.intermediate_output = {}
        self.results = {}
        self.start_time = {}
        self.tournament_history = {}
        self.termination_history = {}
        self.tournaments = {}
        self.read_from = {"conf id":1 , "instance_id":1 , "index":1 }
        self.scenario = scenario

        # todo logging dic should be provided somewhere else -> DOTAC-37
        logging.basicConfig(filename=f'./selector/logs/{self.scenario.log_folder}/Target_Algorithm_Cache.logger', level=logging.INFO,
                            format='%(asctime)s %(message)s')

    def put_intermediate_output(self, conf_id, instance_id, value):
        logging.info(f"Getting intermediate_output: {conf_id}, {instance_id}, {value} ")

        if conf_id not in self.intermediate_output:
            self.intermediate_output[conf_id] = {}

        if instance_id not in self.intermediate_output[conf_id]:
            self.intermediate_output[conf_id][instance_id] = [value]
        else:
            self.intermediate_output[conf_id][instance_id] = self.intermediate_output[conf_id][instance_id] + [value]

    def get_intermediate_output(self):
        # TODO store from where we have read last and contiue form there
        return self.intermediate_output

    def put_result(self, conf_id, instance_id, result):
        logging.info(f"Getting final result: {conf_id}, {instance_id}, {result}")
        if conf_id not in self.results:
            self.results[conf_id] = {}

        if instance_id not in self.results[conf_id]:
            self.results[conf_id][instance_id] = result

    def get_results(self):
        logging.info(f"Publishing results")
        return self.results

    def get_results_single(self, conf_id, instance_id):
        result = False
        if conf_id in list(self.results.keys()):
            if instance_id in list(self.results[conf_id].keys()):
                result = self.results[conf_id][instance_id]
        return result

    def put_start(self,conf_id, instance_id, start):
        logging.info(f"Getting start: {conf_id}, {instance_id}, {start} ")
        if conf_id not in self.start_time:
            self.start_time[conf_id] = {}

        if instance_id not in self.start_time[conf_id]:
            self.start_time[conf_id][instance_id] = start

    def get_start(self):
        logging.info(f"Publishing start")
        return self.start_time

    def put_tournament_history(self, tournament):
        self.tournament_history[tournament.id ] = tournament

    def get_tournament_history(self):
        return self.tournament_history

    def put_tournament_update(self, tournament):
        self.tournaments[tournament.id] = tournament

    def remove_tournament(self,tournament):
        self.tournaments.pop(tournament.id)

    def get_tournament(self):
        return list(self.tournaments.values())

    def put_termination_history(self, conf_id, instance_id):
        if conf_id not in self.termination_history:
            self.termination_history[conf_id] = []

        if instance_id not in self.termination_history[conf_id]:
            self.termination_history[conf_id].append(instance_id)
        else:
            logging.info("This should not happen: we kill something we already killed")

    def get_termination_history(self):
        return self.termination_history

    def get_termination_single(self, conf_id, instance_id):
        termination = False
        if conf_id in list(self.termination_history.keys()):
            if instance_id in list(self.termination_history[conf_id]):
                termination = True
        return termination

    def save_rt_results(self):
        with open(f"./selector/logs/{self.scenario.log_folder}/run_history.json", 'a') as f:
            history = {str(k):v for k,v in self.results.items()}
            json.dump(history, f, indent=2)

    def save_tournament_history(self):
        with open(f"./selector/logs/{self.scenario.log_folder}/tournament_history.json", 'a') as f:
            history = {str(k): v for k, v in self.tournament_history.items()}
            json.dump(history, f, indent=4, cls=TournamentEncoder)

    def save_tournament_history(self):
        with open(f"./selector/logs/{self.scenario.log_folder}/tournament_history.json", 'a') as f:
            history = {str(k): v for k, v in self.tournament_history.items()}
            json.dump(history, f, indent=4, cls=TournamentEncoder)



"""This module contains functions for selection of points."""
import numpy as np
from selector.hp_point_selection import select_point


class PointSelector:
    """Generic point selector class."""

    def __init__(self, features=None):
        """Initialize class."""
        self.selection_history = {}
        self.features = features

    def select_points(self, pool, number_of_points, iteration):
        """Generic point selector method."""
        pass


class RandomSelector(PointSelector):
    """Random point selector class."""

    def __init__(self):
        """Initialize class."""
        super().__init__()

    def select_points(self, pool, number_of_points, iteration, seed=False):
        """
        Randomly select a subset of configurations from the pool to run.

        :param pool: dic. Pool of configurations to select from
        :param number_of_points: int. Number of points to select from the pool.
        :param iteration: int. Iteration identifier which stores the selection
                          for later reference
        :return: list. Ids of configurations from pool that are selected
        """
        if seed:
            np.random.seed(seed)
        selected_points = np.random.choice(list(pool), number_of_points,
                                           replace=False)
        self.selection_history[iteration] = selected_points

        return selected_points.tolist()


class HyperparameterizedSelector(PointSelector):
    u"""
    Hyperparameterized selection of generated points.

    Based on:
    Carlos Ansótegui, Meinolf Sellmann, Tapan Shah,
    Kevin Tierney,
    Learning to Optimize Black-Box Functions With
    Extreme Limits on the Number of Function Evaluations,
    2021, International Conference on Learning and Intelligent
    Optimization, 7-24
    """

    def __init__(self):
        """Initialize class."""
        super().__init__()

    def select_points(self, scenario, pool, number_of_points, epoch,
                      max_epoch, features, weights, results, max_evals=100,
                      seed=False):
        """
        Select configurations subset from pool based on scoring function.

        :param pool: dic. Pool of configurations to select from
        :param number_of_points: int. Number of points to select from the pool.
        :param iteration: int. Iteration identifier which stores the selection
                          for later reference
        :param MAX_EPOCH: int. How many simulations per selecte point
        :return: list. Ids of configurations from pool that are selected
        """
        selected_points = select_point(scenario, list(pool), max_evals,
                                       number_of_points, pool, epoch,
                                       max_epoch, features, weights, seed)

        self.selection_history[epoch] = selected_points

        return selected_points



"""This module contains the point generation class."""

import uuid
import random
from selector.generators.random_point_generator import random_point


class PointGen:
    """Interface for point generation."""

    def __init__(self, scenario, gm=random_point, seed=False):
        """
        Initialize PointGen.

        : param scenario: scenario object
        : param gm: point generating method to use
        """
        self.s = scenario
        self.gen_method = gm
        self.seed = seed

    def point_generator(self, **kwargs):
        """
        Running point generation according to object setting.

        : param meta: meta data a point generator requires
        : **kwargs: depend on gen_method
        return: configuration/point generated
        """
        if self.seed:
            self.id = uuid.UUID(int=random.getrandbits(self.seed))
        else:
            self.id = uuid.uuid4()
        configuration = self.gen_method(self.s, self.id, **kwargs)

        return configuration



"""This module contains functions for the CPPL surrogate."""
import os
import copy
import sys
from scipy.stats import norm
from sklearn.preprocessing import OneHotEncoder
import uuid
import numpy as np
from scipy.linalg import sqrtm
import scipy
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from selector.pool import Configuration, Generator, ParamType
from selector.point_gen import PointGen
from selector.generators.random_point_generator import random_point
from selector.generators.default_point_generator import (
    check_no_goods,
    check_conditionals,
    default_point
)
from selector.generators.random_point_generator import reset_conditionals

np.set_printoptions(threshold=sys.maxsize)
sys.path.append(os.getcwd())

from threadpoolctl import ThreadpoolController
controller = ThreadpoolController()


class CPPL:
    """Surrogate from CPPL."""

    def __init__(self, scenario, seed, features, pool_size=192, alpha=1,
                 gamma=0.1, w=0.1, random_prob=0.2, mutation_prob=0.8,
                 pca_dimension_configurations=8, pca_dimension_instances=8,
                 model_update="Batch", v_hat_norm=None, theta_norm="zero_one",
                 feature_normaliser="max", ensemble=True):
        """
        CPPL Initializition.

        Implementation based on the paper: "Pool-based realtime algorithm
        configuration: A preselection bandit approach"
        """
        self.scenario = scenario
        self.features = features
        self.seed = seed
        self.ensemble = ensemble

        self.v_hat_norm = v_hat_norm
        self.theta_norm = theta_norm
        self.feature_normaliser = feature_normaliser

        self.pool_size = pool_size
        self.pool = [random_point(scenario, uuid.uuid4())
                     for _ in range(self.pool_size - 1)] + \
                    [default_point(scenario, uuid.uuid4())]
        self.feature_store = {}

        self.model_update = model_update
        self.alpha = alpha
        self.gamma = gamma
        self.w = w
        self.t = 0

        self.random_prob = random_prob
        self.mutation_prob = mutation_prob
        self.number_new_confs = 3

        self.pca_dimension_configurations = pca_dimension_configurations
        if len(scenario.parameter) < self.pca_dimension_configurations:
            self.pca_dimension_configurations = len(scenario.parameter)

        self.pca_dimension_instances = pca_dimension_instances
        if len(list(self.features.values())[0]) < self.pca_dimension_instances:
            self.pca_dimension_instances = len(list(self.features.values())[0])
        elif len(self.features) < self.pca_dimension_instances:
            self.pca_dimension_instances = len(self.features)

        self.pca_number_confs_calibration = 100

        self.context_dim = self.pca_dimension_configurations * 2 + \
            self.pca_dimension_instances * 2 + \
            self.pca_dimension_configurations * self.pca_dimension_instances

        self.theta_hat = np.random.random_sample(self.context_dim)
        self.theta_bar = copy.copy(self.theta_hat)

        self.gradient_sum = {0: np.zeros((self.context_dim, self.context_dim))}
        self.hessian_sum = {0: np.zeros((self.context_dim, self.context_dim))}

        self.process_parameter()

        instance_feature_matrix = np.array(list(self.features.values()))
        self.instance_feature_standard_scaler = StandardScaler()
        transformed_features = \
            self.instance_feature_standard_scaler.\
            fit_transform(instance_feature_matrix)
        for instance, counter in zip(self.features.keys(),
                                     range(len(self.features.keys()))):
            self.features[instance] = transformed_features[counter]

        self.calibrate_pca()
        self.best_q = None
        self.identity_store = {}
        self.confidences = {}

    def process_parameter(self):
        """
        Figure out the parameter types and get values/calibrate encoding and scaling
        """

        cat_params = []
        cont_int_params = []
        # figure out the param type
        for param in self.scenario.parameter:
            if param.type == ParamType.categorical:
                cat_params.append(param)
            else:
                cont_int_params.append(param)

        # treat cont and int params
        self.lower_b_params = np.zeros(len(cont_int_params))
        self.upper_b_params = np.zeros(len(cont_int_params))

        for i in range(len(cont_int_params)):
            bound = cont_int_params[i].bound
            self.lower_b_params[i] = float(bound[0])
            self.upper_b_params[i] = float(bound[-1])

        # init one hot encoding for cat params:
        list_of_bounds = []
        list_o_of_default = []
        # With longest this is some super wired hack I have to do for the OneHotEncoder
        # If I do not find tha parameter with the longest string to parse in as default
        # the categories will become dtype Ux with x being the lenght of the longest string in the default
        # if then there are values that are longer then UX which get parsed in later these will be cut to the length..
        for param in cat_params:
            if isinstance(param.default, (bool, np.bool_)):
                list_of_bounds.append(list(map(str, list(map(int, param.bound)))))
                list_o_of_default.append(str(int(param.default)))
            else:
                longest = 0
                list_of_bounds.append(param.bound)
                for bound in param.bound:
                    if len(bound) > longest:
                        longest = len(bound)
                        cp = bound
                list_o_of_default.append(cp)

        self.cat_params_names = [p.name for p in cat_params]
        self.cont_int_params_names = [p.name for p in cont_int_params]

        if len(self.cat_params_names) > 0:
            self.o_h_enc = OneHotEncoder(categories=list_of_bounds)
            self.o_h_enc.fit(np.array(list_o_of_default).reshape(1, -1) )

    def scale_conf(self, configuration):
        """
        Scale and encode a conf. Cont/int parameter are scaled between 0 and 1. Cat. parameter are one hot encodede
        :param configuration: Configuraion
        :return:
        """
        cat_params_on_conf = np.zeros(len(self.cat_params_names), dtype=object)
        cont_int_params_of_conf = np.zeros(len(self.cont_int_params_names), dtype=float)

        for param, value in configuration.conf.items():
            if param in self.cat_params_names:
                if isinstance(value, (bool, np.bool_)):
                    cat_params_on_conf[self.cat_params_names.index(param)] = str(int(value))
                else:
                    cat_params_on_conf[self.cat_params_names.index(param)] = value
            else:
                cont_int_params_of_conf[self.cont_int_params_names .index(param)] = value

        cont_int_scaled = (cont_int_params_of_conf - self.lower_b_params) / (self.upper_b_params - self.lower_b_params).reshape(1, -1)

        if len(self.cat_params_names) > 0:
            cat_params_on_conf = self.o_h_enc.transform(cat_params_on_conf.reshape(1, -1)).toarray()

        return np.concatenate((cont_int_scaled, cat_params_on_conf), axis=None)

    def calibrate_pca(self):
        """
        Calibarte the PCA and Scalers for the featuremap
        """

        random_generator = PointGen(self.scenario, random_point)
        if len(self.cat_params_names) > 0:
            para_vector_size = len([item for sublist in self.o_h_enc.categories_ for item in sublist]) + len(self.cont_int_params_names)
        else:
            para_vector_size = len(self.cont_int_params_names)

        conf_matrix = np.zeros((self.pca_number_confs_calibration, para_vector_size))
        for i in range(self.pca_number_confs_calibration):
            point = random_generator.point_generator()
            conf_matrix[i] = self.scale_conf(point)

        self.pca_configurations = PCA(n_components=self.pca_dimension_configurations)
        pca_conf = self.pca_configurations.fit_transform(conf_matrix)

        self.pca_instances = PCA(n_components=self.pca_dimension_instances)
        pca_features = self.pca_instances.fit_transform(np.array(list(self.features.values())))

        sample_size = min(pca_conf.shape[0], pca_features.shape[0])
        pca_conf = pca_conf[:sample_size]
        pca_features = pca_features[:sample_size]

        feature_map_matrix = np.zeros((sample_size * sample_size, self.context_dim))

        fc = 0
        for i in range(sample_size):
            for j in range(sample_size):
                feature_map_matrix[fc] = np.concatenate((pca_conf[i], pca_features[j], pca_conf[i] ** 2, pca_features[j] ** 2,
                                                         (pca_conf[i] * pca_features[j].reshape(-1,1)).flatten())).flatten()
                fc = fc + 1

        self.feature_map_scaler = StandardScaler()
        self.feature_map_scaler.fit(feature_map_matrix)

    def compute_feature_map(self, conf, instance_features):
        """
        For a conf/instance pair compute the quadratic feature map and scale
        :param conf: Configuration object from selector
        :param instance_features: Np array of instances features
        :return:
        """

        conf_values = self.scale_conf(conf)

        conf_values = self.pca_configurations.transform(conf_values.reshape(1, -1))
        instance_features = self.pca_instances.transform(instance_features.reshape(1, -1))

        features = np.concatenate((conf_values, instance_features, conf_values**2, instance_features ** 2,
                                   (conf_values * instance_features.reshape(-1,1)).flatten().reshape(1,-1)),axis=1).flatten()

        # features = features /max(features)
        if self.feature_normaliser == "max":
            features = features / max(features)
        elif self.feature_normaliser == "zero_one":
            features = (features - min(features)) / (max(features) - min(features ))

        return features

    def to_log_space(self):
        pass

    def from_log_space(self):
        pass

    def update_feature_store(self, conf, instance):
        """
        For a conf/instance pair compute the features and store them in a feature store for latere
        :param conf: Configuration object from selector
        :param instance: Instance ID
        :return:
        """
        if conf.id not in self.feature_store:
            self.feature_store[conf.id] = {}

        if instance not in self.feature_store[conf.id]:
            self.feature_store[conf.id][instance] = self.compute_feature_map(conf, self.features[instance])

    def compute_a(self, theta, tried_conf_ids, instance_id):
        """
        Compute a of the paper
        """
        sum = 0

        for conf in tried_conf_ids:
            sum = sum + self.feature_store[conf][instance_id] * \
                np.exp(np.dot(self.feature_store[conf][instance_id], theta))
        return sum

    def compute_b(self, theta, tried_conf_ids, instance_id):
        """
        Compute b of the paper
        """
        sum = 0
        for conf in tried_conf_ids:
            sum = sum + np.exp(
                np.dot(self.feature_store[conf][instance_id], theta))
        return sum

    def compute_c(self, theta, tried_conf_ids, instance_id):
        """
        Compute c of the paper
        """
        sum = 0
        for conf in tried_conf_ids:
            sum = sum + np.exp(np.dot(self.feature_store[conf][instance_id], theta)) * \
                np.outer(self.feature_store[conf][instance_id], self.feature_store[conf][instance_id])
        return sum

    def compute_gradient(self, theta, winner_id, tried_confs, instance_id):
        """
        Compute the gradient for a given feedback
        """
        a = self.compute_a(theta, tried_confs, instance_id)
        b = self.compute_b(theta, tried_confs, instance_id)

        return self.feature_store[winner_id][instance_id] - (a / b)

    def compute_hessian(self,theta, winner_id, tried_confs, instance_id):
        """
        Compute the hessian for a given feedback
        """
        a = self.compute_a(theta, tried_confs, instance_id)
        b = self.compute_b(theta, tried_confs, instance_id)
        c = self.compute_c(theta, tried_confs, instance_id)
        return (np.outer(a, a) / b ** 2) - (c / b)

    def update_running_sums(self, theta, winner_id, tried_confs, instance_id):
        """
        Update the rolling sums of gradients and hessian for a feedback
        """

        if self.t not in self.gradient_sum.keys():
            gradient = self.compute_gradient(theta, winner_id, tried_confs, instance_id)
            self.gradient_sum[self.t] = self.gradient_sum[self.t - 1] + np.outer(gradient, gradient)

        if self.t not in self.hessian_sum.keys():
            hessian = self.compute_hessian(theta, winner_id, tried_confs, instance_id)
            self.hessian_sum[self.t] = self.hessian_sum[self.t - 1] + hessian ** 2

    def compute_confidence(self, theta, instance_id, conf):
        """
        Compute the confidence for an instance/conf combination
        """
        # Control How many threads/cores numpy and scipy use
        @controller.wrap(limits=self.scenario.tournament_size, user_api='openmp')
        @controller.wrap(limits=self.scenario.tournament_size, user_api='blas')
        def compute_confidence_threaded(theta, instance_id, conf):
            v = (1 / self.t) * self.gradient_sum[self.t]

            s = (1 / self.t) * self.hessian_sum[self.t]

            try:
                s_inv = np.linalg.inv(s)
            except:
                s_inv = np.linalg.pinv(s)

            sigma = (1 / self.t) * np.dot(np.dot(s_inv, v), s_inv)

            M = np.exp(2 * np.dot(self.feature_store[conf][instance_id], theta)) * \
                np.outer(self.feature_store[conf][instance_id], self.feature_store[conf][instance_id])

            sigma_root = sqrtm(sigma)
            I_hat = np.linalg.norm(np.dot(np.dot(sigma_root, M), sigma_root))

            return self.w * np.sqrt((2 * np.log(self.t) + self.context_dim + 2 * np.sqrt((self.context_dim * np.log(self.t)))) * I_hat)

        return compute_confidence_threaded(theta, instance_id, conf)

    def update_model_single_observation(self, winner_id, tried_confs, instance_id):
        """
        Update the thetas of the model for a single instance feedback
        """
        grad = self.compute_gradient(self.theta_hat, winner_id, tried_confs, instance_id)
        self.theta_hat = self.theta_hat + self.gamma * self.t ** (-self.alpha) * grad

        if self.theta_norm == "max":
            self.theta_hat = self.theta_hat / max(self.theta_hat)
        elif self.theta_norm == "zero_one":
            self.theta_hat = (self.theta_hat - min(self.theta_hat)) / (max(self.theta_hat ) - min(self.theta_hat ))
        #self.theta_hat = self.theta_hat/max(self.theta_hat)
        self.theta_bar = ((self.t - 1) * (self.theta_bar)) / self.t + (self.theta_hat / self.t)

    def update_model_mini_batch(self, winner_ids, tried_confs, instance_ids):
        """
        Update the thetas of the model for feedback over multiple instances
        """
        grad_mean = np.zeros(self.context_dim)

        for uc in range(len(winner_ids)):
            grad = self.compute_gradient(self.theta_hat, winner_ids[uc], tried_confs[uc], instance_ids[uc])
            grad_mean = grad_mean + grad

        grad_mean = grad_mean / len(winner_ids)
        self.theta_hat = self.theta_hat + self.gamma * self.t ** (-self.alpha) * grad_mean

        self.theta_hat = self.theta_hat / max(self.theta_hat)

        self.theta_bar = ((self.t - 1) * (self.theta_bar)) / self.t + (self.theta_hat / self.t)

    def select_from_set(self, conf_set, instance_set, n_to_select):
        """
        For a set of configurations and instances select the most promising configurations
        """

        v_hat = np.zeros((len(conf_set), len(instance_set)))
        confidence = np.zeros((len(conf_set), len(instance_set)))
        for i, conf in enumerate(conf_set):
            for next_instance, _ in enumerate(instance_set):
                v_hat[i][next_instance] = np.exp(np.inner(self.feature_store[conf.id][instance_set[next_instance]], self.theta_bar))

                if self.t > 0:
                    confidence[i][next_instance] = self.compute_confidence(self.theta_bar, instance_set[next_instance], conf.id)

        v_hat_s = v_hat.sum(axis=1)

        if self.v_hat_norm == "max":
            v_hat_s = v_hat_s / max(v_hat_s)
        elif self.v_hat_norm == "zero_one":
            v_hat_s = (v_hat_s - min(v_hat_s)) / (max(v_hat_s) - min(v_hat_s))

        if self.t > 0:
            confidence_s = confidence.sum(axis=1)
            if self.v_hat_norm == "max":
                confidence_s = confidence_s / max(confidence_s)
            elif self.v_hat_norm == "zero_one":
                confidence_s = (confidence_s - min(confidence_s)) / (max(confidence_s) - min(confidence_s))
        else:
            confidence_s = np.zeros(v_hat_s.shape)

        quality = v_hat_s + confidence_s

        qual_sort = (-quality).argsort()

        selection = qual_sort[:n_to_select]
        return [conf_set[i] for i in selection], [[v_hat_s[i], confidence_s[i]] for i in qual_sort]

    def delete_from_pool(self, instance_set):
        """
        Based on the feedback delete poor performing configurations from the pool
        """
        v_hat = np.zeros((self.pool_size, len(instance_set)))
        confidence = np.zeros((self.pool_size, len(instance_set)))

        for i, conf in enumerate(self.pool):
            for next_instance, _ in enumerate(instance_set):
                v_hat[i][next_instance] = np.exp(np.inner(self.feature_store[conf.id][instance_set[next_instance]], self.theta_bar))

                if self.t > 0:
                    confidence[i][next_instance] = self.compute_confidence(self.theta_bar, instance_set[next_instance], conf.id)

        v_hat_s = v_hat.sum(axis=1)

        if self.v_hat_norm == "max":
            v_hat_s = v_hat_s / max(v_hat_s)
        elif self.v_hat_norm == "zero_one":
            v_hat_s = (v_hat_s - min(v_hat_s)) / (max(v_hat_s) - min(v_hat_s))

        if self.t > 0:
            confidence_s = confidence.sum(axis=1)
            if self.v_hat_norm == "max":
                confidence_s = confidence_s / max(confidence_s)
            elif self.v_hat_norm == "zero_one":
                confidence_s = (confidence_s - min(confidence_s)) / (max(confidence_s) - min(confidence_s))
        else:
            confidence_s = np.zeros(v_hat_s.shape)

        discard_index = []
        for c in range(self.pool_size):
            for oc in range(self.pool_size):
                if c != oc and v_hat_s[oc] - confidence_s[oc] > v_hat_s[c] + confidence_s[c]:
                    discard_index.append(c)
                    break
        dis = []
        for i in sorted(discard_index, reverse=True):
            dis.append(self.pool[i])
            del self.pool[i]

    def create_new_conf(self, parent_one, parent_two):
        """
        Create new configurations based on the genetic procedure described
        """
        no_good = True
        while no_good:
            new_conf = {}
            rn = np.random.uniform()

            if rn < self.random_prob:
                new_conf = random_point(self.scenario, uuid.uuid4())
                new_conf.generator = Generator.cppl
            else:
                for param, setting in parent_one.conf.items():

                    rn = np.random.uniform()
                    if rn < 0.5:
                        new_conf[param] = setting
                    else:
                        new_conf[param] = parent_two.conf[param]
                rn = np.random.uniform()

                if rn < self.mutation_prob:
                    possible_mutations = random_point(self.scenario, uuid.uuid4())
                    param_to_mutate = np.random.choice(list(new_conf.keys()))
                    new_conf[param_to_mutate] = possible_mutations.conf[param_to_mutate]

                identity = uuid.uuid4()

                new_conf = Configuration(identity,
                                         new_conf,
                                         Generator.cppl)

            cond_vio = check_conditionals(self.scenario, new_conf.conf)
            if cond_vio:
                new_conf.conf = reset_conditionals(self.scenario, new_conf.conf, cond_vio)

            no_good = check_no_goods(self.scenario, new_conf.conf)
        return new_conf

    def add_to_pool(self, past_instances):
        """
        Add the most promising newly created configurations to the pool
        """
        number_to_create = self.pool_size - len(self.pool)
        new_promising_conf = []

        if number_to_create > 0:
            if len(self.pool) > 1:
                best_to_confs, _ = self.select_from_set(self.pool, past_instances, 2)
                conf_one, conf_two = best_to_confs[0], best_to_confs[1]
            else:
                conf_one = self.pool[0]
                conf_two = random_point(self.scenario, uuid.uuid4())
                conf_two.generator = Generator.cppl
            possible_new_confs = []
            for nc in range(self.number_new_confs * number_to_create):
                possible_new_confs = possible_new_confs + [self.create_new_conf(conf_one, conf_two)]
            for instance in past_instances:
                for c in possible_new_confs:
                    self.update_feature_store(c, instance)
            # TODO I have to ensure that all the confs here are diffrent...
            new_promising_conf, _ = self.select_from_set(possible_new_confs, past_instances, number_to_create)

        self.pool = self.pool + new_promising_conf

    def update(self, previous_tournament, c, results, terminations, instance_features=None, ac_runtime=None):
        """
        Updated the model with given feedback
        :param results: nested dic containing rt feedback for the conf instance pairs in previous_tournament
        :param previous_tournament: Tournament
        :return:
        """
        if instance_features:
            instance_feature_matrix = np.array(list(instance_features.values()))
            transformed_features = self.instance_feature_standard_scaler.transform(instance_feature_matrix)
            for instance, counter in zip(instance_features.keys(), range(len(instance_features.keys()))):
                self.features[instance] = transformed_features[counter]

        best_conf_store = []
        rest_conf_store = []
        instance_store = []

        confs_w_feedback = previous_tournament.best_finisher + previous_tournament.worst_finisher

        for instance in previous_tournament.instance_set:
            results_on_instance = {}

            for c in self.pool + confs_w_feedback:
                cond_vio = check_conditionals(self.scenario, c.conf)
                if cond_vio:
                    c.conf = reset_conditionals(self.scenario, c.conf, cond_vio)

                self.update_feature_store(c, instance)

            for c in previous_tournament.configuration_ids:
                if not np.isnan(results[c][instance]):
                    results_on_instance[c] = results[c][instance]
                else:
                    # OMIT the capped data in model update
                    if c in terminations:
                        if instance in terminations[c]:
                            continue
                    else:
                        # This conf/inst pair was a time limit reach
                        results_on_instance[c] = self.scenario.cutoff_time

            if self.best_q is None:
                if len(results_on_instance.values()) == 0:
                    import sys
                    self.best_q = sys.maxsize
                else:
                    self.best_q = min(results_on_instance.values())
            elif len(results_on_instance.values()) == 0: 
                if min(results_on_instance.values()) < self.best_q:
                    self.best_q = min(results_on_instance.values())

            best_conf_on_instance = min(results_on_instance, key=results_on_instance.get)

            if not self.ensemble:
                if results_on_instance[best_conf_on_instance] >= self.scenario.cutoff_time:
                    self.pool = [random_point(self.scenario, uuid.uuid4()) for _ in range(self.pool_size -1 )] + [default_point(self.scenario, uuid.uuid4())]
                    for c in self.pool:
                        [self.update_feature_store(c, i) for i in previous_tournament.instance_set]
                    continue

            tried = previous_tournament.configuration_ids.copy()

            # Through suggesting a conf multiple times with diffrent id we may get feedback for the same conf
            # under a diffrent id. Here we map that back in order to update for the same conf.
            if self.ensemble:
                if best_conf_on_instance in self.identity_store:
                    best_conf_on_instance = self.identity_store[best_conf_on_instance]
                    for c in range(len(tried)):
                        if tried[c] in self.identity_store:
                            tried[c] = self.identity_store[tried[c]]

            instance_store.append(instance)
            best_conf_store.append(best_conf_on_instance)
            rest_conf_store.append(tried)

            self.t = self.t + 1

            # this is actually the same as running "batch" with a best_conf_store of len() 1
            if self.model_update == "SGD":
                self.update_model_single_observation(best_conf_on_instance, tried, instance)

            self.update_running_sums(self.theta_bar, best_conf_on_instance, tried, instance)

        if self.model_update == "Batch" and len(best_conf_store) > 0:
            self.update_model_mini_batch(best_conf_store, rest_conf_store, instance_store)

        self.delete_from_pool(previous_tournament.instance_set)

        self.add_to_pool(previous_tournament.instance_set)

    def get_suggestions(self, scenario, n_to_select, d, r, next_instance_set, instance_features=None):
        """
        Suggest configurations to run next based on the instances that are comming
        """

        if instance_features:
            instance_feature_matrix = np.array(list(instance_features.values()))
            transformed_features = self.instance_feature_standard_scaler.transform(instance_feature_matrix)
            for instance, counter in zip(instance_features.keys(), range(len(instance_features.keys()))):
                self.features[instance] = transformed_features[counter]

        for instance in next_instance_set:
            for c in self.pool:
                self.update_feature_store(c, instance)

        suggest, ranking = self.select_from_set(self.pool, next_instance_set, n_to_select)

        suggest = copy.deepcopy(suggest)
        for sugg in suggest:
            sugg.generator = Generator.cppl
            # We have to make sure that the ids of the configurations returend are unique.
            # I.e for two different tournaments we may suggest the same conf twice.
            # In that case the conf needs a unique id.
            # Later update() we need to map that back
            if self.ensemble:
                identity = uuid.uuid4()
                self.identity_store[identity] = sugg.id
                sugg.id = identity

        return suggest, ranking

    def suggest_from_outside_pool(self, conf_set, n_to_select, next_instance_set, instance_features=None):

        if instance_features:
            instance_feature_matrix = np.array(list(instance_features.values()))
            transformed_features = self.instance_feature_standard_scaler.transform(instance_feature_matrix)
            for instance, counter in zip(instance_features.keys(), range(len(instance_features.keys()))):
                self.features[instance] = transformed_features[counter]

        for instance in next_instance_set:
            for c in conf_set:
                self.update_feature_store(c, instance)

        suggest, ranking = self.select_from_set(conf_set, next_instance_set, n_to_select)

        return suggest, ranking

    def predict(self, suggestions, next_instance_set):
        """Predict performance/quality of configurations with GGA++ epm.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :param next_instance_set: list, next instances to run tournament on
        :return [mean, var]: numpy.ndarray, mean and variance of predicted
            performance/quality
        """
        ranking = self.suggest_from_outside_pool(suggestions, len(suggestions),
                                                 next_instance_set)[1]
        v = []
        c = []

        for i, _ in enumerate(ranking):
            v.append(ranking[i][0])
            c.append(ranking[i][1])

        v = np.array(v)
        c = np.array(c)

        self.v = v
        self.c = c

        return v, c

    def probability_improvement(self, suggestions, results, next_instance_set):
        """Compute probability of improvement.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :param next_instance_set: list, next instances to be run
        :return pi: numpy.ndarray, probabilities of improvement
        """
        v = self.v
        c = self.c

        std = np.sqrt(c)

        pi = norm.cdf((self.best_q - v) / std)
        pi_output = []
        for p in pi:
            pi_output.append([p])

        return pi_output

    def expected_improvement(self, suggestions, next_instance_set):
        """Compute expected improvement via CPPL model.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :param next_instance_set: list, next instances to be run
        :return ei: numpy.ndarray, expected improvements
        """
        mean = self.v
        var = self.c

        std = np.sqrt(var)

        def calculate_ei():
            z = (self.best_q - mean) / std
            return (self.best_q - mean) * norm.cdf(z) + std * norm.pdf(z)

        if np.any(std == 0.0):
            stdc = np.copy(std)
            std[stdc == 0.0] = 1.0
            ei = calculate_ei()
            ei[stdc == 0.0] = 0.0
            return ei
        else:
            return calculate_ei()



import re
import warnings
from selector.pool import Parameter, ParamType

import numpy as np

boolean_yes = ["on", "yes", "true"]
boolean_no = ["no", "off", "false"]
boolean_options = boolean_no + boolean_yes


def get_ta_arguments_from_pcs(para_file):
    """
     Read in a file that contains the target algorithm parameters. The file is .pcs and adheres to the structure of
     AClib.
    :param para_file: str. Path to the .pcs file
    :return: list, list. A list containing information on the parameters of the target algorithm & a list containing
    parameter value assignments that are not possible
    """

    no_goods = []
    parameters = []
    conditionals = {}

    with open(para_file, 'r') as pf:
        for line in pf:
            line = line.strip().split("#", 1)[0]

            # skip empty lines
            if line == "":
                continue

            line_split = line.split(None, 1)
            param_name = line_split[0].replace(" ", "")
            param_info = line_split[1]  # TODO This will error if forbidden params do not have any spaces

            if "|" not in param_info:

                # cat
                if '{' in param_info:
                    param_type, bounds, defaults, original_bound = get_categorical(param_name, param_info)
                    if param_type is not None:
                        parameters.append(Parameter(param_name, param_type, bounds, defaults, {}, '', original_bound))
                # forbidden
                elif '{' in param_name:
                    no_good = get_no_goods(line, parameters)
                    no_goods.append(no_good)
                # cont.
                elif '[' in param_info:
                    param_type, bounds, defaults, scale = get_continuous(param_name, param_info)
                    parameters.append(Parameter(param_name, param_type, bounds, defaults, {}, scale, []))

            # conditionals
            elif '|' in param_info:
                condition_param, condition = get_conditional(param_name, param_info, parameters)

                if param_name not in conditionals:
                    conditionals[param_name] = {condition_param: condition}
                else:
                    conditionals[param_name].update({condition_param: condition})

            else:
                raise ValueError(f"The parameter file {para_file} contains unreadable elements. Check that"
                                 f" the structure adheres to AClib")

    # adding conditionals to parameters
    for pc in conditionals:
        condition_found = False
        for parameter in parameters:
            if re.search(r'\b' + str(pc) + r'\b', parameter.name):
                parameter.condition.update(conditionals[pc])
                condition_found = True
        # This should only be a warning: We may have conditions for cat. parameters that are not configurable.
        # We ignore these.
        if not condition_found:
            warnings.warn(f"Condition {pc}|{conditionals[pc]} will be dropped since either {pc} is "
                          f"not configurable or does not exist")

    return parameters, no_goods, conditionals


def get_categorical(param_name, param_info):
    """
    For a categorical parameter: check if its parsed attributes are valid and extract information on the parameter
    :param param_name: name of the parameter
    :param param_info: raw parameter information
    :return: param_type , bounds, defaults of the parameter
    """

    bounds = re.search(r'\{(.*)\}', param_info).group().strip("{ }").split(",")
    bounds = [b.replace(" ","") for b in bounds]
    defaults = re.findall(r'\[(.*)\]*]', param_info)
    original_bound = []

    # When len(bounds) ==1: The parameter can not be configured. Later we have to also raise a warning that conditions
                                                                # on it are ignored
    if len(bounds) == 1:
        warnings.warn(f"For parameter {param_name} bounds of length 1 were passed. The parameter will "
                      f"be ignored for configuration.")

        param_type, bounds, defaults = None, None, None
    elif bounds[0] in boolean_options and bounds[1] in boolean_options:
        param_type = ParamType.categorical
        original_bound = bounds

        if defaults[0] in boolean_yes:
            defaults = True
        elif defaults[0] in boolean_no:
            defaults = False
        else:
            raise ValueError(f"For parameter {param_name} the parsed defaults are not within [yes, no, on, off]")
        bounds = [b in boolean_yes for b in bounds]
        bounds = sorted(bounds)

    elif isinstance(str(bounds[0]), str) & isinstance(str(defaults[0]), str):
        param_type = ParamType.categorical
        defaults = str(defaults[0])
        bounds = [str(b).replace(" ", "") for b in bounds]
        if defaults not in bounds:
            raise ValueError(f"For parameter {param_name} the default value is not within the range of the bounds")

    else:
        raise ValueError(f"For parameter {param_name} the parsed bounds were not boolean or categorical")

    return param_type, bounds, defaults, original_bound


def get_continuous(param_name, param_info):
    """
    For a continuous parameter: check if its parsed attributes are valid and extract information on the parameter
    :param param_name: name of the parameter
    :param param_info: raw parameter information
    :return: param_type , bounds, defaults,scale of the parameter
    """

    scale = re.search(r'[a-zA-Z]+', param_info)
    param_info = re.findall(r'\[[^\]]*]', param_info)
    bounds = param_info[0].strip("[] ").split(",")
    defaults = param_info[1].strip("[] ")

    # checking for set scale
    if scale and "i" in scale.group():
        param_type = ParamType.integer
        scale = scale.group().strip("i")

        if isinstance(int(defaults), int):
            defaults = int(defaults)
        else:
            raise ValueError(f"For parameter {param_name} the parsed defaults are not integer")
        bounds = [int(b) for b in bounds]

    else:
        param_type = ParamType.continuous

        if isinstance(float(defaults), float):
            defaults = float(defaults)
        else:
            raise ValueError(f"For parameter {param_name} the parsed defaults are not continuous")
        bounds = [float(b) for b in bounds]

        if scale is None:
            scale = ''
        else:
            scale = scale.group()

    if not bounds[0] <= defaults <= bounds[1]:
        raise ValueError(f"For parameter {param_name} the default value is not within the range of the bounds")

    return param_type, bounds, defaults, scale


def get_conditional(param_name, param_info, parameters):
    """
    For a parameter: get the information on conditionals
    :param param_name: name of the parameter
    :param param_info: raw parameter information
    :return: condition_param, condition
    """
    param_info = param_info.strip(" | ")

    condition = re.search(r'\{(.*)\}', param_info).group().strip("{ }").split(",")
    condition_param = re.search(r'.+?(?= in)', param_info).group().replace(" ", "")

    for p in parameters:
        if p.name == condition_param:
            p_type = p.type

    if condition[0] in boolean_options:
        condition = [c in boolean_yes for c in condition]
    elif p_type == ParamType.categorical:
        condition = [str(c).strip(" ") for c in condition]
    elif p_type == ParamType.continuous:
        condition = [float(c) for c in condition]
    elif p_type == ParamType.integer:
        condition = [int(c) for c in condition]
    else:
        raise ValueError(f"For parameter {param_name} the parsed conditions could not be read")

    return condition_param, condition


def get_no_goods(no_good, parameters):
    """
    Takes an string of form: {param_1=value_1 , param_2=value_2, ...} and returns a dic of the no good
    :param no_good: str. Takes an string of form: {param_1=value_1 , param_2=value_2, ...}
    :return: dic
    """

    forbidden = {}
    no_good = no_good.strip("{ }").split(",")

    for ng in no_good:
        param, value = ng.split("=")
        param = param.strip()
        value = value.strip()

        for p in parameters:
            if p.name == param:
                p_type = p.type

        if value in boolean_yes:
            value = True
        elif value in boolean_no:
            value = False
        elif p_type == ParamType.continuous:
            value = float(value)
        elif p_type == ParamType.integer:
            value = int(value)
        elif p_type == ParamType.categorical:
            value = str(value)
        else:
            raise ValueError(f"For no good {no_good} the parameter values are not known")

        forbidden[param] = value

    return forbidden


def read_instance_paths(instance_set_path):
    """
    Read in instances from an AClib instance file
    :param instance_set_path: str. Path to the instance file
    :return: list. List of paths to the instances
    """

    instance_set = []

    with open(instance_set_path, 'r') as f:
        for line in f:
            instance_set.append(line.strip())

    seen = set()
    uniq = []
    for i in instance_set:
        if i not in seen:
            uniq.append(i)
            seen.add(i)
        else:
            warnings.warn(f"Instance {i} is not unique in the train set")

    return instance_set


def read_instance_features(feature_set_path):
    """
    Read in features from an AClib features file
    :param feature_set_path: str. Path to the feature file
    :return: dic, list. Dic with the read in features list with the feature names
    """

    features = {}
    with open(feature_set_path, 'r') as f:
        lines = f.readlines()
        feature_names = lines[0].strip().split(",")[1:]

        for line in lines[1:]:
            line = line.strip().split(",")
            if line[0] != "":
                features[line[0]] = np.array(line[1:], dtype=np.single)

    return features, feature_names



import os
import shutil
from datetime import datetime
import dataclasses
import json
import uuid
from enum import Enum
import ray
import numpy as np

def clear_logs(folder_for_run = None):
    """
    Clear the logs
    """
    if folder_for_run == None:
        folder_for_run = "latest"

    for folder in [f'./selector/logs/{folder_for_run}' ,f'./selector/logs/{folder_for_run}/ta_logs']:
        for filename in os.listdir(folder):
            file_path = os.path.join(folder, filename)
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)

def check_log_folder(folder_for_run = None):
    if folder_for_run == None:
        folder_for_run = "latest"
    if not os.path.exists("./selector/logs"):
        os.makedirs("./selector/logs")

    if not os.path.exists(f'./selector/logs/{folder_for_run}'):
        os.makedirs(f'./selector/logs/{folder_for_run}')

    if not os.path.exists(f'./selector/logs/{folder_for_run}/ta_logs'):
        os.makedirs(f'./selector/logs/{folder_for_run}/ta_logs')

def save_latest_logs(folder_for_run):
    if folder_for_run == "latest":
        shutil.copytree('./selector/logs/latest', f"./selector/logs/{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}")

def log_termination_setting(logger, scenario):
    if scenario.termination_criterion == "total_runtime":
        logger.info(f"The termination criterion is: {scenario.termination_criterion}")
        logger.info(f"The total runtime is: {scenario.wallclock_limit}")
    elif scenario.termination_criterion == "total_tournament_number":
        print(scenario.termination_criterion)
        logger.info(f"The termination criterion is: {scenario.termination_criterion}")
        logger.info(f"The total number of tournaments is: {scenario.total_tournament_number}")
    else:
        logger.info(f"No valid termination criterion has been parsed. "
                    f"The termination criterion will be set to runtime.")
        logger.info(f"The total runtime is: {scenario.wallclock_limit}")

class TournamentEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            data_dic = dataclasses.asdict(o)
            if "ray_object_store" in data_dic.keys():
                del data_dic["ray_object_store"]
            return data_dic
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, Enum):
            return str(o)
        elif isinstance(o, ray._raylet.ObjectRef):
            return str(o)
        elif isinstance(o, np.bool_):
            return bool(o)
        elif isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, dict):
            for k in o.keys():
                if isinstance(k, uuid.UUID):
                    o[str(k)] = o.pop(k)
            return o

        return super().default(o)

class ConfEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        else:
            return super().default(obj)



from dataclasses import dataclass
from enum import Enum


@dataclass
class Configuration:
    id: int
    conf: dict
    generator: Enum


@dataclass
class Parameter:
    name: str
    type: str
    bound: list
    default: int
    condition: list
    scale: str
    original_bound: list


@dataclass
class Tournament:
    id: int
    best_finisher: list
    worst_finisher: list
    configurations: list
    configuration_ids: list
    ray_object_store: dict
    instance_set: list
    instance_set_id: int


class ParamType(Enum):
    categorical = 1
    continuous = 2
    integer = 3


class TaskType(Enum):
    target_algorithm = 1
    monitor = 2


class Generator(Enum):
    default = 1
    random = 2
    var_graph = 3
    lhc = 4
    smac = 5
    ggapp = 6
    cppl = 7


class Status(Enum):
    win = 1
    cap = 2
    timeout = 3
    stop = 4
    running = 5


class Surrogates(Enum):
    SMAC = 1
    GGApp = 2
    CPPL = 3



"""This module contains functions for the SMAC surrogate."""

from smac.configspace import ConfigurationSpace
from smac.optimizer.epm_configuration_chooser import EPMChooser
from smac.scenario.scenario import Scenario
from smac.configspace import Configuration
from ConfigSpace.conditions import InCondition, AndConjunction
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter
)
from ConfigSpace.forbidden import (
    ForbiddenEqualsClause,
    ForbiddenAndConjunction
)
from smac.stats.stats import Stats
from smac.runhistory.runhistory import RunHistory
from smac.runhistory.runhistory2epm import RunHistory2EPM4LogCost
from smac.epm.rf_with_instances import RandomForestWithInstances
from smac.epm.util_funcs import get_types
from smac.optimizer.acquisition import EI, PI
from smac.optimizer.ei_optimization import LocalSearch
from numpy.random import RandomState
from smac.tae import StatusType
import numpy
import uuid
import random
import copy
import math
import sys

from selector.pool import ParamType, Generator
from selector.pool import Configuration as SelConfig
from selector.generators.default_point_generator import (
    check_conditionals,
    check_no_goods
)
from selector.generators.random_point_generator import (
    reset_no_goods,
    random_set_conf
)
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler

from smac.configspace import convert_configurations_to_array
from typing import List, Optional, Tuple
import itertools

from threadpoolctl import ThreadpoolController
controller = ThreadpoolController()


#  Adjust SMAC's LocalSearch to only suggest as many points as we actually ask
class LocalSearch(LocalSearch):
    """Implementation of SMAC's local search."""

    def _get_initial_points(
        self,
        num_points: int,
        runhistory: RunHistory,
        additional_start_points: Optional[List[Tuple[float, Configuration]]],
    ) -> List[Configuration]:

        if runhistory.empty():
            init_points = \
                self.config_space.sample_configuration(size=num_points)
        else:
            # initiate local search
            configs_previous_runs = runhistory.get_all_configs()

            # configurations with the highest previous EI
            configs_previous_runs_sorted = \
                self._sort_configs_by_acq_value(configs_previous_runs)
            configs_previous_runs_sorted = \
                [conf[1] for conf in configs_previous_runs_sorted[:num_points]]

            # configurations with the lowest predictive cost,
            # check for None to make unit tests work
            if self.acquisition_function.model is not None:
                conf_array = \
                    convert_configurations_to_array(configs_previous_runs)
                costs = self.acquisition_function.model.\
                    predict_marginalized_over_instances(conf_array)[0]
                assert len(conf_array) == len(costs), (conf_array.shape,
                                                       costs.shape)

                # In case of the predictive model returning the prediction
                # for more than one objective per configuration
                # (for example multi-objective or EIPS) it is not immediately
                # clear how to sort according to the cost
                # of a configuration. Therefore, we simply follow the ParEGO#
                # approach and use a random scalarization.
                if len(costs.shape) == 2 and costs.shape[1] > 1:
                    weights = numpy.array([self.rng.rand()
                                          for _ in range(costs.shape[1])])
                    weights = weights / numpy.sum(weights)
                    costs = costs @ weights

                # From here
                # http://stackoverflow.com/questions/20197990/how-to-make-argsort-result-to-be-random-between-equal-values
                random = self.rng.rand(len(costs))
                # Last column is primary sort key!
                indices = numpy.lexsort((random.flatten(), costs.flatten()))

                # Cannot use zip here because the indices array cannot index
                # the rand_configs list, because the second is a
                # pure python list
                configs_previous_runs_sorted_by_cost = \
                    [configs_previous_runs[ind] 
                     for ind in indices][:num_points]
            else:
                configs_previous_runs_sorted_by_cost = []

            if additional_start_points is not None:
                additional_start_points = \
                    [asp[1] for asp in additional_start_points[:num_points]]
            else:
                additional_start_points = []

            init_points = []
            init_points_as_set = set()  # type: Set[Configuration]
            for cand in itertools.chain(
                configs_previous_runs_sorted,
                configs_previous_runs_sorted_by_cost,
                additional_start_points,
            ):
                if cand not in init_points_as_set:
                    init_points.append(cand)
                    init_points_as_set.add(cand)

            # Here actual adjustment to give only num_points initial points.
            # We remove math.floor(remove num_pomits / 2)
            # configs_previous_runs_sorted, getting 50/50 from
            # configs_previous_runs_sorted and 
            # configs_previous_runs_sorted_by_cost but favor the former,
            # if odd number.
            if len(init_points) > num_points:
                remove_nr = math.floor((len(init_points) - num_points))
                if remove_nr > math.floor(num_points / 2):
                    remove_nr = math.floor(num_points / 2)
                for rr in range(remove_nr):
                    del init_points[num_points - 1 - rr]
                if len(init_points) > num_points:
                    init_points = init_points[:num_points]

        return init_points


class SmacSurr():
    """Surrogate from SMAC."""

    def __init__(self, scenario, seed=False, pca_dim=8):
        """Initialize smac surrogate.

        :param scenario: dict, scenario translated for smac
        :param seed: int, random seed
        :param pca_dim: int, PCA on instance features (default same as GGApp)
        """
        if not seed:
            self.seed = False
        else:
            self.seed = seed

        # Control How many threads/cores numpy and scipy use
        @controller.wrap(limits=scenario.tournament_size,
                         user_api='openmp')
        @controller.wrap(limits=scenario.tournament_size,
                         user_api='blas')
        def threaded_init(scenario, pca_dim):
            self.pca_dim = pca_dim
            self.best_val = sys.maxsize
            self.param_bounds = {}
            for param in scenario.parameter:
                self.param_bounds[param.name] = param.bound
            self.stats = Stats
            self.rs = RandomState
            self.s = copy.deepcopy(scenario)
            inst_feats = numpy.array(list(scenario.features.values()))
            self.inst_ids = list(scenario.features.keys())
            self.selector_scenario = scenario
            scaler = MinMaxScaler()
            inst_feats = scaler.fit_transform(inst_feats)
            inst_feats = numpy.nan_to_num(inst_feats)
            pca = PCA(n_components=self.pca_dim)
            pca_inst_feats = pca.fit_transform(inst_feats)
            self.pca_inst_feats, self.inst_ids = \
                self.pca_inst_feat_file(self.selector_scenario.instance_file,
                                        self.selector_scenario.feature_file,
                                        self.inst_ids, pca_inst_feats)
            self.scenario, self.config_space, self.types, self.bounds \
                = self.transfom_selector_scenario_for_smac(self.selector_scenario)
            self.rh = RunHistory(overwrite_existing_runs=True)
            self.rh2epm = RunHistory2EPM4LogCost(scenario=self.scenario,
                                                 num_params=len(self.config_space),
                                                 success_states=StatusType)
            self.rafo \
                = RandomForestWithInstances(configspace=self.config_space,
                                            types=self.types,
                                            bounds=self.bounds,
                                            seed=self.seed,
                                            instance_features=self.pca_inst_feats,
                                            num_trees=10,  # Same as in GGApp
                                            pca_components=self.pca_dim)
            self.aaf = EI(model=self.rafo)
            self.aafpi = PI(model=self.rafo)
            self.afm = LocalSearch(acquisition_function=self.aaf,
                                   config_space=self.config_space,
                                   max_steps=10,
                                   n_steps_plateau_walk=1,
                                   vectorization_min_obtain=2,
                                   vectorization_max_obtain=64
                                   )
            self.surr = EPMChooser(scenario=self.scenario,
                                   stats=self.stats,
                                   runhistory=self.rh,
                                   runhistory2epm=self.rh2epm,
                                   model=self.rafo,
                                   acq_optimizer=self.afm,
                                   acquisition_func=self.aaf,
                                   rng=self.rs,
                                   random_configuration_chooser=None,
                                   predict_x_best=False)

        threaded_init(scenario, pca_dim)

    def transfom_selector_scenario_for_smac(self, scenario):
        """Transform scenario to SMAC formulation.

        :param scenario: scenario object from selector
        :return scenario: scenario object from SMAC
        """
        config_space = ConfigurationSpace()
        types = []
        bounds = []
        self.neg_cat = {}

        # Setup parameter space.
        for param in scenario.parameter:
            if param.scale:
                log = True
            else:
                log = False

            if param.type == ParamType.integer:
                if param.bound[0] < 0:
                    log = False
                parameter \
                    = UniformIntegerHyperparameter(param.name,
                                                   param.bound[0],
                                                   param.bound[1],
                                                   default_value=param.default,
                                                   log=log)

                config_space.add_hyperparameter(parameter)

            elif param.type == ParamType.continuous:
                if param.bound[0] < 0:
                    log = False
                parameter \
                    = UniformFloatHyperparameter(param.name,
                                                 param.bound[0],
                                                 param.bound[1],
                                                 default_value=param.default,
                                                 log=log)

                config_space.add_hyperparameter(parameter)

            elif param.type == ParamType.categorical:
                if type(param.bound[0]) is bool:
                    bounds = []
                    for pb in param.bound:
                        if pb is True:
                            bounds.append(True)
                        elif pb is False:
                            bounds.append(False)
                    if param.default is True:
                        default = bounds.index(True)
                    else:
                        default = bounds.index(True)
                else:
                    if param.bound[0].replace('-', '').isdigit():
                        bounds = []
                        # adjust neg categorical parameters for smac surr
                        if '-' in param.bound[0]:
                            add = -1 * int(param.bound[0])
                            bounds = [str(int(i) + add) for i in param.bound]
                            default = str(int(param.default) + add)
                            self.neg_cat[param.name] = add
                        else:
                            for pb in param.bound:
                                bounds.append(str(pb))
                            default = str(param.default)
                    elif '.' in param.bound[0]:
                        if param.bound[0].replace('-', '').replace('.', '')\
                                .isdigit():
                            bounds = []
                            # adjust neg categorical parameters for smac surr
                            if '-' in param.bound[0]:
                                add = -1 * float(param.bound[0])
                                bounds = [str(float(i) + add)
                                          for i in param.bound]
                                default = str(float(param.default) + add)
                                self.neg_cat[param.name] = add
                            else:
                                for pb in param.bound:
                                    bounds.append(str(pb))
                                default = str(param.default)
                    else:
                        bounds = []
                        for pb in param.bound:
                            bounds.append(str(pb))
                        default = str(param.default)

                parameter \
                    = CategoricalHyperparameter(param.name,
                                                bounds,
                                                default_value=default)

                config_space.add_hyperparameter(parameter)

        cond_list = []

        def transform_conditionals(config_space, condvalues):
            if isinstance(config_space[parent],
                          CategoricalHyperparameter):
                if type(condvalues[0]) == str:
                    if condvalues[0].replace('-', '').isdigit():
                        for i, condval in enumerate(condvalues):
                            condvalues[i] = str(condval)
                    elif '.' in condvalues[0]:
                        if condvalues[0].replace('-', '').\
                                replace('.', '').isdigit():
                            for i, condval in enumerate(condvalues):
                                condvalues[i] = str(condval)
                    else:
                        for i, condval in enumerate(condvalues):
                            condvalues[i] = condval

            return condvalues

        # Set up conditionals.
        for child, parents in scenario.conditionals.items():
            if len(parents) >= 2:
                conjunction = []
                for parent, vals in parents.items():
                    condvalues = vals
                    for condval in condvalues:
                        if condval == 'True':
                            condval = True
                        elif condval == 'False':
                            condval = False
                    # adjust neg categorical parameters for smac surr
                    if parent in self.neg_cat:
                        for i, c in enumerate(condvalues):
                            if '.' in c:
                                condvalues[i] = str(float(c) +
                                                    self.neg_cat[parent])
                            else:
                                condvalues[i] = str(int(c) +
                                                    self.neg_cat[parent])

                    condvalues = \
                        transform_conditionals(config_space, condvalues)

                    # Conjunction needed in ConfigSpace if parameter
                    # has more than one conditionals
                    conjunction.append(
                        InCondition(child=config_space[child],
                                    parent=config_space[parent],
                                    values=condvalues))

                cond_list.append(AndConjunction(*conjunction))

            else:
                parent = list(parents.keys())[0]
                vals = list(parents.values())[0]
                condvalues = vals
                for condval in condvalues:
                    if condval == 'True':
                        condval = True
                    elif condval == 'False':
                        condval = False
                # adjust neg categorical parameters for smac surr
                if parent in self.neg_cat:
                    for i, c in enumerate(condvalues):
                        if '.' in c:
                            condvalues[i] = str(float(c) +
                                                self.neg_cat[parent])
                        else:
                            condvalues[i] = str(int(c) + self.neg_cat[parent])

                condvalues = \
                    transform_conditionals(config_space, condvalues)

                cond_list.append(
                    InCondition(child=config_space[child],
                                parent=config_space[parent],
                                values=condvalues))

            config_space.add_conditions(cond_list)

        # Setup no goods.
        for ng in scenario.no_goods:
            ng_list = []
            for param, val in ng.items():

                if isinstance(config_space[param],
                              CategoricalHyperparameter):
                    if type(val) == str:
                        if val.replace('-', '').isdigit():
                            val = str(val)
                        elif '.' in val:
                            if val.replace('-', '').\
                                    replace('.', '').isdigit():
                                val = str(val)
                        else:
                            val = val
                        # adjust neg categorical parameters for smac surr
                        if param in self.neg_cat:
                            if '.' in val:
                                val = str(float(val) + self.neg_cat[param])
                            else:
                                val = str(int(val) + self.neg_cat[param])

                ng_list.append(ForbiddenEqualsClause(config_space[param], val))

            config_space.add_forbidden_clause(
                ForbiddenAndConjunction(*ng_list))

        types, bounds = get_types(config_space)

        self.cutoff_time = scenario.cutoff_time

        # SMAC scenario object
        s = Scenario({'run_obj': scenario.run_obj,
                      'cutoff': scenario.cutoff_time,
                      'runcount-limit': 10,
                      'cs': config_space,
                      'deterministic': True,
                      'acq_opt_challengers': scenario.tournament_size,
                      'instance_file': scenario.instance_file,
                      'feature-file': scenario.feature_file})

        return s, config_space, types, bounds

    def transform_values(self, conf, pred=False):
        """Transform configuration values in SMAC format.

        :param conf: object, selector.pool.Configuration
        :return config: dict, parameter values
        """
        config = {}

        # Check conditionals and reset parameters if violated
        cond_vio = check_conditionals(self.s, conf.conf)

        for cv in cond_vio:
            conf.conf[cv] = None

        for param in self.config_space:
            if param in conf.conf:
                if conf.conf[param] is None:
                    config[param] = None
                    continue
            else:
                config[param] = None
                continue

            if isinstance(self.config_space[param],
                          UniformFloatHyperparameter):
                config[param] = float(conf.conf[param])

            if isinstance(self.config_space[param],
                          UniformIntegerHyperparameter):
                config[param] = int(conf.conf[param])

            if isinstance(self.config_space[param],
                          CategoricalHyperparameter):
                if isinstance(conf.conf[param], (numpy.bool_, bool)):
                    config[param] = bool(conf.conf[param])
                else:
                    if type(conf.conf[param]) == str or \
                            isinstance(conf.conf[param], numpy.str_):
                        if conf.conf[param].replace('-', '').isdigit():
                            config[param] = str(conf.conf[param])
                        elif '.' in conf.conf[param]:
                            if conf.conf[param].replace('-', '').\
                                    replace('.', '').isdigit():
                                config[param] = str(conf.conf[param])
                        else:
                            if pred:
                                config[param] = \
                                    str(self.param_bounds[param].
                                        index(conf.conf[param]))
                            else:
                                config[param] = conf.conf[param]
                    else:
                        if type(conf.conf[param]) == int or \
                                type(conf.conf[param]) == float:
                            config[param] = conf.conf[param]
                        else:
                            config[param] = str(conf.conf[param])
                    if param in self.neg_cat:
                        if type(config[param]) == float:
                            config[param] = config[param] + self.neg_cat[param]
                        elif type(config[param]) == int:
                            config[param] = config[param] + self.neg_cat[param]
                        elif '.' in config[param]:
                            config[param] = str(float(config[param]) +
                                                self.neg_cat[param])
                        else:
                            config[param] = str(int(config[param]) +
                                                self.neg_cat[param])

        return config

    def update(self, history, configs, results, terminations, ac_runtime=None):
        """Update SMAC epm.

        :param history: Tournament history
        :param conf: object, selector.pool.Configuration
        :param state: object selector.pool.Status, status of this point
        :param tourn_nr: int, number of tournament, which to update with
        """
        config_dict = {}
        for c in configs:
            config_dict[c.id] = c
        # instances in tournament
        instances = history.instance_set

        if ac_runtime >= self.selector_scenario.wallclock_limit * 0.15 and \
                (len(instances) * len(config_dict)) * 2 \
                < len(self.surr.runhistory.data):
            ds = (len(instances) * len(config_dict))
            to_delete = []
            delete_by_ids = []
            for rd in self.surr.runhistory.data.keys():
                if (rd not in
                        to_delete and len(delete_by_ids) < ds) or \
                        (rd not in to_delete and rd.config_id in 
                            delete_by_ids):
                    to_delete.append(rd)
                    if rd.config_id not in \
                            delete_by_ids and \
                            len(delete_by_ids) < ds:
                        delete_by_ids.append(rd.config_id)

            for td in to_delete:
                del self.surr.runhistory.data[td]

            for td in delete_by_ids:
                dbc = self.surr.runhistory.ids_config[td]
                del self.surr.runhistory.ids_config[td]
                del self.surr.runhistory.config_ids[dbc]

        for cid in config_dict.keys():
            # config in results
            for ins in instances:
                # OMIT every censored date in update
                if cid in terminations:
                    if ins in terminations[cid]:
                        continue
                conf = config_dict[cid]

                if ins in results[cid]:
                    if not numpy.isnan(results[cid][ins]):
                        state = StatusType.SUCCESS

                    else:
                        # This conf/inst pair was a time limit reach
                        state = StatusType.TIMEOUT
                        results[cid][ins] = self.cutoff_time

                    config = self.transform_values(conf)
                    config = dict(sorted(zip(config.keys(),
                                             config.values())))
                    # adjust neg categorical parameters for smac surr
                    for c, v in config.items():
                        if c in self.neg_cat:
                            if type(self.neg_cat[c]) is int:
                                config[c] = str(v)
                            else:
                                config[c] = str(v)

                    config = Configuration(self.config_space, values=config)

                    self.surr.runhistory.add(config, results[cid][ins],
                                             results[cid][ins], state,
                                             ins, self.seed)

    def get_suggestions(self, scenario, n_samples, *args):
        """Get point suggestions from SMAC.

        :param scenario: scenario object from selector
        :param n_samples: int, number of point suggestions to generate
        :return suggestions: list, list of configurations
        """
        suggestions = []
        added = 0
        param_order = []
        params = scenario.parameter
        for p in params:
            param_order.append(p.name)

        import time

        start = time.time()

        # Tell SMAC how many suggetions to make
        self.surr.scenario.acq_opt_challengers = n_samples

        # Tell SMAC features of the next instances
        if args[2] is not None:
            next_inst_feats = []
            for inst in args[2]:
                next_inst_feats.append(self.pca_inst_feats[self.inst_ids.index(inst)])
            self.surr.model.instance_features = numpy.array(next_inst_feats)
        
        while len(suggestions) < n_samples:
            sugg = self.surr.choose_next()
            sugg = list(sugg)

            for s in sugg:
                if added < n_samples:
                    if not self.seed:
                        identity = uuid.uuid4()
                    else:
                        identity = uuid.UUID(int=random.getrandbits(self.seed))

                    sugg_items = s.get_dictionary()
                    config_setting = {}
                    for po in param_order:
                        if po in sugg_items:
                            config_setting[po] = sugg_items[po]

                    # adjust neg categorical parameters for target algorithms
                    for k, v in config_setting.items():
                        if k in self.neg_cat:
                            if '.' in v:  # for floats
                                config_setting[k] = str(float(v) -
                                                        self.neg_cat[k])
                            else:  # for ints
                                config_setting[k] = str(int(v) -
                                                        self.neg_cat[k])

                    # For consistency, set random value for turned off
                    # parameters (due to conditionals), since SMAC accounts
                    # for conditionals in config generation and all other
                    # generators do not
                    for param in self.s.parameter:
                        if param.name in self.s.conditionals and \
                                param.name not in config_setting:
                            config_setting[param.name] = \
                                random_set_conf([param])[param.name]

                    suggestions.append(
                        SelConfig(identity,
                                  config_setting,
                                  Generator.smac))

                    # Check no goods and reset values if violated
                    ng_vio = check_no_goods(scenario, suggestions[added].conf)
                    while ng_vio:
                        suggestions[added].conf = \
                            reset_no_goods(scenario, suggestions[added].conf)
                        ng_vio = check_no_goods(scenario,
                                                suggestions[added].conf)

                    added += 1

        return suggestions

    def transform_for_epm(self, confs, pred=False):
        """Transform configuration to suit SMAC epm.

        :param confs: list of objects, [selector.pool.Configuration,]
        :param pred: bool, True if used in predict function
        :return configs: list, transformed configurations
        """
        configs = []

        for con in confs:
            config = self.transform_values(con, pred)

            for i, c in config.items():
                if c is None:
                    config[i] = numpy.nan

            configs.append(list(config.values()))

        configs = numpy.array(configs, dtype=float)

        return configs

    def predict(self, confs, inst):
        """Predict performance/quality of configurations with SMAC epm.

        :param confs: list of objects, [selector.pool.Configuration,]
        :return [mean, var]: numpy.ndarray, mean and variance of predicted
            performance/quality
        """
        all_configs = self.transform_for_epm(confs, pred=True)

        if not any(isinstance(i, list) for i in all_configs):
            all_configs = [all_configs]

        if self.surr.model.rf is not None:

            m = []
            v = []

            m_av = 0
            v_av = 0

            n_samples, _, _ = self.surr._collect_data_to_train_model()
            
            for c in all_configs[0]:
                if n_samples.shape[0] > self.pca_dim:
                    for infe in inst:
                        instfeat = numpy.asarray(
                            self.pca_inst_feats[self.inst_ids.index(infe)])
                        mean, var = \
                            self.surr.model._predict(
                                X=numpy.array([numpy.append(c, instfeat)]))
                        for i, val in enumerate(mean):
                            m_av += mean[i][0]
                            v_av += var[i][0]

                        m.append(m_av / len(inst))
                        v.append(v_av / len(inst))
                else:
                    mean, var = \
                        self.surr.model._predict(X=numpy.array([c]))
                    for i, val in enumerate(mean):
                        m += mean[i][0]
                        v += var[i][0]

            return numpy.array(m), numpy.array(v)

        else:

            return None

    def expected_improvement(self, suggestions, _):
        """Compute expected improvement via SMAC epm.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :return ei: numpy.ndarray, expected improvements
        """
        configs = self.transform_for_epm(suggestions)

        if self.surr.model.rf is not None:
            self.surr.acquisition_func.update(eta=self.best_val,
                                              model=self.surr.model)
            ei = self.surr.acquisition_func._compute(X=configs)
            expimp = []

            for e in ei:
                expimp.append(list(e)[0])

            return expimp
        else:
            return [[0] for s in suggestions]

    def probability_improvement(self, suggestions, results, i):
        """Compute probability of improvement.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :return pi: numpy.ndarray, probabilities of improvement
        """
        if len(results.values()) == 0:
            import sys
            self.best_val = sys.maxsize
        else:
            self.best_val = min(min(list(d.values()))
                                for d in list(results.values()))

        self.aafpi.update(eta=self.best_val, model=self.surr.model)

        if self.aafpi.eta is not None and \
                self.surr.model.rf is not None:
            configs = self.transform_for_epm(suggestions)

            pi = self.aafpi._compute(X=configs)

            probimp = []

            for p in pi:
                probimp.append(list(p))

            return probimp
        else:
            return [[0] for _ in suggestions]

    def pca_inst_feat_file(self, train_insts, feature_file, insts, pca_feats):
        """Generate instance feature file with PCA features.

        :param train_insts: str, path to training instance file
        :param feature_file: str, path to original feature file
        :param insts: list, instance names as strings
        :param pca_feats: instance features after PCA
        """
        training_instances = []
        with open(f'{train_insts}', 'r') as f:
            for line in f:
                training_instances.append(line.strip())

        with open(f'{feature_file}', 'r') as f:
            feature_names = f.readline()

        self.selector_scenario.feature_file = \
            './selector/logs/' + self.selector_scenario.log_folder + \
            '/features_PCA.txt'

        with open(self.selector_scenario.feature_file, 'w') as f:
            f.write(feature_names)

        pca_inst_feats = []
        inst_ids = []

        with open(self.selector_scenario.feature_file, 'a') as f:
            for name, feats in zip(insts, pca_feats):
                if name in training_instances:
                    pca_inst_feats.append(feats)
                    inst_ids.append(name)
                    feat_string = ",".join(feats.astype('str'))
                    f.write(f"{name}, {feat_string}\n")

        return numpy.array(pca_inst_feats), inst_ids



"""
This module contains the default point generator.

It also contains functions to check validity of the configuration setting.
"""

from selector.pool import Configuration, ParamType, Generator


def check_conditionals(s, config_setting):
    """
    Check if conditionals are violated.

    : param s: scenario object
    : param config_setting: configuration setting to be checked
    return: List of parameter names to turn off
    """
    cond_vio = []
    for condition in s.conditionals:
        for cond in s.conditionals[condition]:
            for param in s.parameter:
                if param.name == cond:
                    if cond in config_setting:
                        if config_setting[cond] is not None:
                            if type(s.conditionals[condition][cond][0]) == int:
                                config_setting[cond] = \
                                    int(config_setting[cond])
                            if type(s.conditionals[condition][cond][0]) \
                                    == float:
                                config_setting[cond] = \
                                    float(config_setting[cond])
                            if param.type == ParamType.categorical and \
                                    (config_setting[cond] not in
                                     s.conditionals[condition][cond] and
                                     condition not in cond_vio and
                                     condition in config_setting):
                                cond_vio.append(condition)
                            elif (param.type == ParamType.continuous or
                                    param.type == ParamType.integer) and \
                                    (config_setting[cond] >
                                     s.conditionals[condition][cond][1] or
                                     config_setting[cond] <
                                     s.conditionals[condition][cond][0]) and \
                                    condition not in cond_vio and \
                                    condition in config_setting:
                                cond_vio.append(condition)

    return cond_vio


def check_no_goods(s, config_setting):
    """
    Check if no goods are violated.

    : param s: scenario object
    : param config_setting: configuration setting to be checked
    return: True/False, if no goods are violated
    """
    check = False
    for ng in s.no_goods:
        params = list(ng.keys())

        if all([i in config_setting for i in params]):
            check = all([ng[i] == config_setting[i] for i in params])
            if check:
                return check

    return check


def default_point(s, identity):
    """
    Default parameter setting is generated in Configuration format.

    : param s: scenario object
    : param identity: uuid to identify configuration
    return: default configuration
    """
    config_setting = {}

    # Generate configuration with default values
    for param in s.parameter:
        config_setting[param.name] = param.default

    # Fill Configuration class with ID and parameter values
    configuration = Configuration(identity, config_setting, Generator.default)

    return configuration



import numpy as np
import math


class InstanceSet:
    def __init__(self, instance_set, start_instance_size, set_size=None, instance_increment_size=None, target_reach=None):
        """

        :param instance_set: set of instances available.
        :param start_instance_size: size of the first instances se to be created
        :param set_size: If not set the biggest instance set to be created includes all instances in instance_set. If
        set the biggest instance set will be of the size of the given int.
        """
        self.instance_set = instance_set
        self.start_instance_size = start_instance_size
        self.instance_sets = []
        self.subset_counter = 0

        if set_size and set_size <= len(instance_set):
            self.set_size = set_size
        else:
            self.set_size = len(instance_set)

        if instance_increment_size:
            self.instance_increment_size = instance_increment_size
        else:
            self.instance_increment_size = self.set_size / np.floor(self.set_size/self.start_instance_size)

        if target_reach:
            set_size = self.start_instance_size
            counter = 0
            while set_size <= self.set_size:
                set_size += self.instance_increment_size
                counter += 1
            self.target_increment = np.ceil(target_reach / counter)
        else:
            self.target_increment = 1

    def next_set(self):
        """
        Create a new instance set with instances not included in any created set before. This is not thread safe and
        should only be called from the master node.
        :param id: int. Id of the new instance set. If it is the first set id will be 1.
        """
        # Get instances that were already chosen
        if self.subset_counter == 0:
            seen_instances = []
        else:
            seen_instances = self.instance_sets[self.subset_counter - 1]

        # If we have not chosen any instance before we create a first set
        if not self.instance_sets:
            if self.start_instance_size > len(self.instance_set):
                raise ValueError("The number of instances provided is smaller than the initial instance set size")
            else:
                new_subset = np.random.choice(self.instance_set, self.start_instance_size, replace=False).tolist()
                self.instance_sets.append(new_subset)
        # If we have seen instances before and are still allowed to choose instances we do so
        elif len(seen_instances) <= self.set_size:
            # We either sample as many instances as the slope tells us or the last remainder set to full instance sice
            number_to_sample = int(min(self.instance_increment_size, self.set_size - len(seen_instances)))
            # We can only select instances not chosen before
            possible_instances = [i for i in self.instance_set if i not in seen_instances]
            new_subset = np.random.choice(possible_instances, number_to_sample, replace=False).tolist()
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1] + new_subset)
        # In case we have chosen all instances but still tournaments to run we use the full set of instances for all
        # following tournaments
        else:
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1])

        self.subset_counter += 1

    def get_subset(self, next_tournament_set_id):
        """
        Create an instance set for the next tournament. The set contains all instances that were included in the
        previous sets as well as a new subset of instances.
        :param next_tournament_set_id: int. Id of the subset to get the next instances for .
        :return: id of the instances set, list containing instances and previous instances of the subset
        """
        assert next_tournament_set_id <= self.subset_counter
        # If we have already created the required subset we return it
        if next_tournament_set_id in range(len(self.instance_sets)):
            next_set = self.instance_sets[next_tournament_set_id]

        elif self.target_increment != 1 and next_tournament_set_id % self.target_increment != 0:
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1])
            next_set = self.instance_sets[next_tournament_set_id - 1]
            self.subset_counter += 1

        # In case we have not, we create the next instance subset
        else:
            self.next_set()
            next_set = self.instance_sets[next_tournament_set_id]
        return next_tournament_set_id, next_set


class FlexInstanceSet(InstanceSet):
    def __init__(self, instance_set, start_instance_size, set_size=None, instance_increment_size=None, target_reach=None, target_start=None):
        """

        :param instance_set: set of instances available.
        :param start_instance_size: size of the first instances se to be created
        :param set_size: If not set the biggest instance set to be created includes all instances in instance_set. If
        set the biggest instance set will be of the size of the given int.
        """
        self.instance_set = instance_set
        self.start_instance_size = start_instance_size
        self.instance_sets = []
        self.subset_counter = 0
        self.target_start = target_start

        if set_size and set_size <= len(instance_set):
            self.set_size = set_size
        else:
            self.set_size = len(instance_set)

        if instance_increment_size:
            self.instance_increment_size = instance_increment_size
        else:
            self.instance_increment_size = self.set_size / np.floor(self.set_size / self.start_instance_size)

        if target_reach:
            set_size = self.start_instance_size
            counter = 0
            while set_size <= self.set_size:
                set_size += self.instance_increment_size
                counter += 1
            if target_start:
                self.target_increment = np.ceil(((target_reach - target_start) / self.set_size))
            else:
                self.target_increment = np.ceil(target_reach / counter)
        else:
            self.target_increment = 1

    def next_set(self):
        """
        Create a new instance set with instances not included in any created set before. This is not thread safe and
        should only be called from the master node.
        :param id: int. Id of the new instance set. If it is the first set id will be 1.
        """
        # Get instances that were already chosen
        if self.subset_counter == 0:
            seen_instances = []
        else:
            seen_instances = self.instance_sets[self.subset_counter - 1]

        # If we have not chosen any instance before we create a first set
        if not self.instance_sets:
            if self.start_instance_size > len(self.instance_set):
                raise ValueError("The number of instances provided is smaller than the initial instance set size")
            else:
                new_subset = np.random.choice(self.instance_set, self.start_instance_size, replace=False).tolist()
                self.instance_sets.append(new_subset)
        # If we have seen instances before and are still allowed to choose instances we do so
        elif len(seen_instances) <= self.set_size:
            # We either sample as many instances as the slope tells us or the last remainder set to full instance sice
            number_to_sample = int(min(self.instance_increment_size, self.set_size - len(seen_instances)))
            # We can only select instances not chosen before
            possible_instances = [i for i in self.instance_set if i not in seen_instances]
            if self.target_start:
                if self.subset_counter >= self.target_start:
                    new_subset = np.random.choice(possible_instances, number_to_sample, replace=False).tolist()
                else:
                    new_subset = []
            else:
                new_subset = np.random.choice(possible_instances, number_to_sample, replace=False).tolist()
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1] + new_subset)
        # In case we have chosen all instances but still tournaments to run we use the full set of instances for all
        # following tournaments
        else:
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1])

        self.subset_counter += 1


class TimedInstanceSet(InstanceSet):
    def __init__(self, instance_set, start_instance_size, set_size=None, runtime=4, start_time=0.1, end_time=0.7):
        """

        :param instance_set: set of instances available.
        :param start_instance_size: size of the first instances se to be created
        :param set_size: If not set the biggest instance set to be created includes all instances in instance_set. If
        set the biggest instance set will be of the size of the given int.
        """
        self.instance_set = instance_set
        self.start_instance_size = start_instance_size
        self.instance_sets = []
        self.subset_counter = 0
        self.runtime = runtime
        self.start_time = start_time
        self.end_time = end_time
        self.instance_increment_size = 0
        self.target_increment = 1

        if set_size and set_size <= len(instance_set):
            self.set_size = set_size
        else:
            self.set_size = len(instance_set)

    def next_set(self):
        """
        Create a new instance set with instances not included in any created set before. This is not thread safe and
        should only be called from the master node.
        :param id: int. Id of the new instance set. If it is the first set id will be 1.
        """
        # Get instances that were already chosen
        if self.subset_counter == 0:
            seen_instances = []
        else:
            seen_instances = self.instance_sets[self.subset_counter - 1]

        # If we have not chosen any instance before we create a first set
        if not self.instance_sets:
            if self.start_instance_size > len(self.instance_set):
                raise ValueError("The number of instances provided is smaller than the initial instance set size")
            else:
                new_subset = np.random.choice(self.instance_set, self.start_instance_size, replace=False).tolist()
                self.instance_sets.append(new_subset)
        # If we have seen instances before and are still allowed to choose instances we do so
        elif len(seen_instances) <= self.set_size:
            # We either sample as many instances as the slope tells us or the last remainder set to full instance sice
            number_to_sample = int(min(self.instance_increment_size, self.set_size - len(seen_instances)))
            # We can only select instances not chosen before
            possible_instances = [i for i in self.instance_set if i not in seen_instances]
            if possible_instances:
                new_subset = np.random.choice(possible_instances, number_to_sample, replace=False).tolist()
                self.instance_sets.append(self.instance_sets[self.subset_counter - 1] + new_subset)
            else:
                self.instance_sets.append(self.instance_sets[self.subset_counter - 1])
        # In case we have chosen all instances but still tournaments to run we use the full set of instances for all
        # following tournaments
        else:
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1])

        self.subset_counter += 1

    def get_subset(self, next_tournament_set_id, time, iteration):
        """
        Create an instance set for the next tournament. The set contains all instances that were included in the
        previous sets as well as a new subset of instances.
        :param next_tournament_set_id: int. Id of the subset to get the next instances for .
        :return: id of the instances set, list containing instances and previous instances of the subset
        """
        assert next_tournament_set_id <= self.subset_counter

        if time >= self.start_time * self.runtime:
            caution = ((1 - (self.set_size - len(self.instance_sets[self.subset_counter - 1])) / self.set_size)) + 1
            time_per_iteration = (time / iteration) * caution
            max_it = max((((self.runtime * self.end_time) - time) / time_per_iteration) * (1 - (time / self.runtime)) * self.end_time, 1)
            if max_it * self.instance_increment_size / self.target_increment < self.set_size - len(self.instance_sets[self.subset_counter - 1]):
                while max_it * self.instance_increment_size / self.target_increment < self.set_size - len(self.instance_sets[self.subset_counter - 1]) and self.set_size != len(self.instance_sets[self.subset_counter - 1]):    
                    if self.target_increment > 1:
                        self.target_increment -= 1
                    elif self.instance_increment_size < int(self.set_size * ((self.set_size - len(self.instance_sets[self.subset_counter - 1])) / max_it)):
                        self.instance_increment_size += 1
                        # Making sure we don't increase more than 2% of complete instance set
                        if self.instance_increment_size > int(self.set_size * 0.02):
                            break
                    else:
                        break
            elif max_it * self.instance_increment_size / self.target_increment >= self.set_size - len(self.instance_sets[self.subset_counter - 1]):
                while max_it * self.instance_increment_size / self.target_increment >= self.set_size - len(self.instance_sets[self.subset_counter - 1]) and self.set_size != len(self.instance_sets[self.subset_counter - 1]):
                    if self.instance_increment_size > 1:
                        self.instance_increment_size -= 1
                    else:
                        self.target_increment += 1

        # If we have already created the required subset we return it
        if next_tournament_set_id in range(len(self.instance_sets)):
            next_set = self.instance_sets[next_tournament_set_id]

        elif self.target_increment != 1 and next_tournament_set_id % self.target_increment != 0:
            self.instance_sets.append(self.instance_sets[self.subset_counter - 1])
            next_set = self.instance_sets[next_tournament_set_id - 1]
            self.subset_counter += 1

        # In case we have not, we create the next instance subset
        else:
            self.next_set()
            next_set = self.instance_sets[next_tournament_set_id]

        return next_tournament_set_id, next_set



"""This module contains point selection functions."""
import numpy as np
import copy
import itertools
from selector.pool import ParamType


def get_relatives(suggested):
    """Get information of relations of suggested points by generator tag.

    :param suggested: list of suggested points
    :return relatives: nested array, indices of related points
    """
    relatives = []
    for s in suggested:
        gen_type = s.generator
        index_list = []
        index_list = [idx for idx, sugg in enumerate(suggested)
                      if sugg != s and sugg.generator == gen_type]
        relatives.append(index_list)

    return np.array(relatives, dtype=object)


def distance_stats(smfeatures, distances):
    """Compute distances statistics.

    :param suggested: list, list of suggested points
    :param distances: list, distance values
    :return smfeatures: array, new features for simulation
    """
    smflen = len(smfeatures[0])
    smfeatures = np.hstack((smfeatures, np.mean(distances, axis=1).reshape(
                            len(distances), 1)))
    smfeatures = np.hstack((smfeatures, np.mean(distances * distances,
                            axis=1).reshape(len(distances), 1)))
    smfeatures = np.hstack((smfeatures, np.std(distances, axis=1).reshape(
                            len(distances), 1)))
    mindist = np.min(distances, axis=1)
    smfeatures = np.hstack((smfeatures, (smfeatures[:, smflen] -
                            mindist).reshape(len(distances), 1)))

    return smfeatures


def simulation(suggested, features, max_evals, selected_points, weights,
               npoints, distances, relatives):
    """Run simulations of config selection.

    :param suggested: list, list of configs/points to select from
    :param features: nested list, features of configs/points
    :param max_eval: int, number of simulation runs per selected point
    :return sfreq: list, how often configs/points were selected in sim
    """
    sugg = list(range(len(suggested)))
    sfreq = np.zeros(len(sugg))

    for evaluation in range(max_evals):
        smsel = copy.copy(selected_points)
        smsugg = copy.copy(sugg)
        smfeatures = copy.copy(features)
        smweights = copy.copy(weights)
        smdistances = copy.copy(distances)

        for selpoint in range(len(selected_points), npoints):

            # After the first point is chosen
            if selpoint > 0:
                # Diversity features to selected points
                simseldist = smdistances[:, smsel]
                smfeatures = distance_stats(smfeatures, simseldist)

            rel_sel = list(itertools.chain.from_iterable(relatives[sel]
                                                         for sel in smsel))
            if rel_sel:
                # Diversity features to selected and related points
                simrelseldist = smdistances[:, rel_sel]
                smfeatures = distance_stats(smfeatures, simrelseldist)

            # Min-max normalization
            minf = np.min(smfeatures, axis=0)
            maxf = np.max(smfeatures, axis=0)
            diff = maxf - minf
            eq = np.where(minf == maxf)[0]
            ge = np.setdiff1d(np.arange(smfeatures.shape[1]), eq,
                              assume_unique=True)
            smfeatures[:, ge] = (smfeatures[:, ge] - minf[ge]) / diff[ge]
            # set no variance features to 0, except for the first
            smfeatures[:, eq[1:]] = 0

            # Probability distribution based on scores
            s_w = 1.0 / (1.0 + np.exp(np.sum(smfeatures *
                         smweights[:, 0:len(smfeatures[0])], axis=1)))

            # Scores based on probability distribution
            scores = np.maximum(0, np.minimum(1, s_w))

            # Select with probability according to scores
            if np.sum(scores) > 0:
                selprob = scores / np.sum(scores)
                selected = np.random.choice(smsugg, 1, p=selprob.tolist())[0]
                selected_idx = smsugg.index(selected)
            else:
                selected = np.random.choice(smsugg, 1)[0]
                selected_idx = smsugg.index(selected)

            # Update frequency of selections
            sfreq[selected] += 1

            # Update point selection within simulation run
            smsel.append(selected_idx)

            # Make sure selected points cannot be selected again in simulation
            del smsugg[selected_idx]
            smfeatures = np.delete(smfeatures, selected_idx, axis=0)
            smweights = np.delete(smweights, selected_idx, axis=0)
            smdistances = np.delete(smdistances, selected_idx, axis=0)
            smfeatures = copy.copy(features[0:len(smfeatures)])

    return sfreq


def normalize_plus_cond_acc(sugg, s):
    """Normalize and account for conditionals.

    :param sugg: list, configuration values
    :param s: scenario
    :return sugg: suggested configuration with normalized and adjusted values
    """
    maximums = {}
    cat_params = []
    if isinstance(s, list):
        psetting = s
    else:
        psetting = s.parameter

    for param in psetting:
        if param.type == ParamType.categorical:
            if len(param.bound) > 2:
                if isinstance(param.bound[0], (str, np.str_)):
                    maximums[param.name] = len(param.bound)
                else:
                    maximums[param.name] = \
                        float(param.bound[len(param.bound) - 1])
            else:
                maximums[param.name] = 1
            cat_params.append(param.name)
        else:
            maximums[param.name] = param.bound[len(param.bound) - 1]

    for point in sugg:
        for key, _ in point.conf.items():
            if key in cat_params:
                if point.conf[key] is True:
                    point.conf[key] = 1
                else:
                    point.conf[key] = 0

    for key, val in maximums.items():
        for point in sugg:
            if key in point.conf:
                if point.conf[key] is None:
                    pass
            elif key in point.conf and maximums[key] > 0:
                point.conf[key] = point.conf[key] / maximums[key]
            elif key in point.conf and maximums[key] < 0:
                point.conf[key] = maximums[key] / point.conf[key]
            elif key not in point.conf:
                point.conf[key] = None

    return sugg


def pairwise_distances(sugg_i, sugg_j):
    """Compute pairwise distances.

    :param sugg_i: list, configuration values
    :param sugg_j: list, configuration values
    :return m: pairwise distances
    """
    m = np.zeros((len(sugg_i), len(sugg_j)))
    for i, s_i in enumerate(sugg_i):
        for j, s_j in enumerate(sugg_j):
            s = 0
            for key in s_i.conf:
                if (s_i.conf[key] is None and s_j.conf[key]
                        is not None) or \
                        (s_i.conf[key] is not None and s_j.conf[key]
                            is None):
                    s = s + 1
                elif (s_i.conf[key] is None and s_j.conf[key]
                        is None):
                    s = s + 0
                else:
                    if isinstance(s_i.conf[key], str) or \
                            isinstance(s_j.conf[key], str):
                        s = s + (float(s_i.conf[key]) - float(s_j.conf[key]))**2
                    else:
                        s = s + (s_i.conf[key] - s_j.conf[key])**2

            m[i, j] = s**0.5

    return np.array(m)


def select_point(scenario, suggested, max_evals, npoints, pool, epoch,
                 max_epoch, features, weights, seed):
    """Generate features and run simultion.

    :param suggested: list, list of configs/points to select from
    :param max_eval: int, number of simulation runs per selected point
    :param npoints: int, number of configs/points requested
    :param pool: list, list of configs/pints to select from
    :param epoch: int, current epoch
    :param max_epoch: int, number of total epochs
    :return selected_points: list, ids of selected configs/points
    """
    if seed:
        np.random.seed(seed)

    relatives = get_relatives(suggested)

    suggested_intact = copy.copy(suggested)

    # Not all points have values for conditional params. In order to
    # compute matching feature vectors, we omit conditional params.
    sugg = copy.deepcopy(suggested)
    sugg = normalize_plus_cond_acc(sugg, scenario)

    distances = pairwise_distances(sugg, sugg)

    selected_points = []
    smselected_points = []

    # Run simulation for every point requested
    for psel in range(npoints):

        sfreq = simulation(suggested, features, max_evals, smselected_points,
                           weights, npoints, distances, relatives)
        sidx = np.argmax(sfreq)
        selected_points.append(suggested_intact[sidx])
        del suggested_intact[sidx]
        smselected_points.append(sidx)
        del suggested[sidx]
        weights = np.delete(weights, sidx, axis=0)
        features = np.delete(features, sidx, axis=0)
        distances = np.delete(distances, sidx, axis=0)

    return selected_points



#!/usr/bin/env python

import pathlib
from setuptools import setup, find_packages


HERE = pathlib.Path(__file__).parent


README = (HERE / "README.md").read_text()

packages = find_packages()
install_requires = [
  line.strip() for line in open('requirements.txt')
]

# This call to setup() does all the work
setup(
    name="selector-ac",
    version="0.1.0.1",
    description="Selector: Ensemble-Based Automated Algorithm Configuration",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/DOTBielefeld/selector",
    author="DOT",
    author_email="dimitri-weiss@web.de",
    packages=find_packages(exclude=["*wrapper*", "test"]),
    license="MIT License",
    python_requires=">=3.8",
    classifiers=[
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX :: Linux",  
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    install_requires=install_requires,
    entry_points={},
)


import ray
import logging
import time
import numpy as np


@ray.remote(num_cpus=1)
class WinnerInstanceMonitor:
    def __init__(self, sleep_time, cache, delta_cap=1):
        """
        Monitor whether the runtime of a configuration on an instance exceeds the worst runtime of a tournament winning
        configuration on that instance multiplied by a constant (delta_cap).
        When a runtime exceeds this bound the configuration/instance pair is terminated.
        The terminated configuration/instance pairs are stored in the termination_history to avoid double killing
        :param sleep_time: Int. Wake up and check whether runtime is exceeded
        :param cache: Ray cache
        :param delta_cap: Int. Constant the current best runtime for each instance is multiplied by
        :return: conf/instance that are killed.
        """
        self.sleep_time = sleep_time
        self.cache = cache
        self.tournaments = []
        self.termination_history = {}
        self.instance_results = {}
        self.delta_cap = delta_cap

        logging.basicConfig(filename=f'./selector/logs/winner_inst_monitor.log', level=logging.INFO,
                            format='%(asctime)s %(message)s')

    def monitor(self):
        logging.info("Starting monitor")

        while True:

            # Get results that are already available for ta runs
            start = time.time()
            results = ray.get(self.cache.get_results.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting results {dur}")

            # get starting times for each conf/instance
            start = time.time()
            start_time = ray.get(self.cache.get_start.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting start {dur}")

            # Get the current tournaments that are in the cache
            start = time.time()
            tournaments = ray.get(self.cache.get_tournament.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting tournaments {dur}")

            # List with the winners of the tournaments currently running
            winners = [t.best_finisher[0] for t in tournaments if t.best_finisher]
            logging.info(f"Current tournament winners: {winners}")

            # We create a dict with the results for the instances for the tournament winners
            # The worst time by a winner for an instance, that did not time out, gets stored in the dict as a benchmark
            # for the instance based capping.
            self.instance_results = {}
            for conf in winners:
                for instance in results[conf.id]:
                    if instance in self.instance_results and results[conf.id][instance] > \
                            self.instance_results[instance] and not np.isnan(results[conf.id][instance]):
                        self.instance_results[instance] = results[conf.id][instance]
                    elif instance not in self.instance_results and not np.isnan(results[conf.id][instance]):
                        self.instance_results[instance] = results[conf.id][instance]

            logging.info(f"Results of the current tournament winners: {self.instance_results}")

            for t in tournaments:
                for conf in t.configurations:
                    instances_conf_finished = []
                    if conf.id in list(results.keys()):
                        instances_conf_finished = list(results[conf.id].keys())

                    instances_conf_planned = list(t.ray_object_store[conf.id].keys())
                    instances_conf_still_runs = [i for i in instances_conf_planned if i not in instances_conf_finished]

                    # We kill a configuration/instance pair, when the runtime exceeds the benchmark runtime
                    # for that instance multiplied by delta_cap
                    for instance in instances_conf_still_runs:
                        if conf.id in start_time and instance in start_time[conf.id] \
                                and instance in self.instance_results:
                            instance_runtime = time.time() - start_time[conf.id][instance]
                            logging.info(
                                f"Monitor kill check: conf.id: {conf.id}, "
                                f"instance: {instance}, "
                                f"instance_runtime: {instance_runtime}, "
                                f"best_instance_runtime: {self.instance_results[instance]}")
                            if instance_runtime > self.instance_results[instance] * self.delta_cap \
                                    and self.termination_check(conf.id, instance):
                                logging.info(
                                    f"Monitor is killing: {conf} {instance} "
                                    f"with id: {t.ray_object_store[conf.id][instance]}")
                                print(f"Monitor is killing: {time.ctime()} {t.ray_object_store[conf.id][instance]}")
                                self.update_termination_history(conf.id, instance)
                                logging.info(f"termination_history: {self.termination_history}")
                                [ray.cancel(t.ray_object_store[conf.id][instance])]
                            else:
                                continue

            time.sleep(self.sleep_time)

    def termination_check(self, conf_id, instance):
        """
        Check if we have killed a conf/instance pair already. Return True if we did not.
        :param conf_id:
        :param instance:
        :return:
        """
        if conf_id not in self.termination_history:
            return True
        elif instance not in self.termination_history[conf_id]:
            return True
        else:
            return False

    def update_termination_history(self, conf_id, instance_id):
        if conf_id not in self.termination_history:
            self.termination_history[conf_id] = []

        if instance_id not in self.termination_history[conf_id]:
            self.termination_history[conf_id].append(instance_id)
        else:
            logging.info(f"This should not happen: we kill something we already killed")



"""This module contains functions for random point generation."""

import numpy as np
import math
import sys
from selector.pool import Configuration, ParamType, Generator
from selector.generators.default_point_generator import (
    check_conditionals,
    check_no_goods
)


def reset_no_goods(s, config_setting):
    """
    Checking for no goods and resetting parameter values if violated.

    : param s: scenario object
    : param config_setting: parameter configuration
    return: config_estting adjusted to no goods
    """
    for ng in s.no_goods:

        violation = True

        if violation:
            ng_values = list(ng.values())
            config_set_values = []

            for ng_element in ng:
                config_set_values.append(config_setting[ng_element])

            if config_set_values == ng_values:
                configs_to_reset = []
                violation = True

                for params in s.parameter:
                    if params.name in ng:
                        configs_to_reset.append(params)

                new_setting = random_set_conf(configs_to_reset)

                for ns in new_setting:
                    config_setting[ns] = new_setting[ns]

            else:
                violation = False
        else:
            continue

    return config_setting


def reset_conditionals(s, config_setting, cond_vio):
    """
    Checking for conditionals and resetting parameter values if violated.

    : param s: scenario object
    : param config_setting: parameter configuration
    : param cond_vio: list of parameters that violate conditionals
    return: config_estting adjusted to no goods
    """
    for cv in cond_vio:
        for ps in s.parameter:
            if cv == ps.name:
                config_setting[cv] = ps.bound[0]

    return config_setting


def random_set_conf(parameter):
    """
    Generating random configuration values for given param space.

    : param parameter: dataclass Parameter, filled out with scenario data
    return: randomly set parameters
    """
    config_setting = {}

    for param in parameter:

        if param.type == ParamType.categorical:
            pv = np.random.choice(param.bound)
            if isinstance(pv, np.bool_):
                config_setting[param.name] = bool(pv)
            else:
                config_setting[param.name] = pv

        elif param.type == ParamType.continuous:
            if param.scale:
                # Generate in logarithmic space
                if param.bound[1] < 0:
                    # Upper and lower bound are negative
                    config_setting[param.name] \
                        = -1 * math.exp(np.random.uniform(
                            low=math.log(param.bound[0] * -1),
                            high=math.log(param.bound[1] * -1)))

                elif param.bound[0] < 0:
                    # Lower bound is negative
                    lb = -1 * math.exp(np.random.uniform(
                        low=math.log(sys.float_info.min),
                        high=math.log(param.bound[0] * -1)))
                    ub = math.exp(np.random.uniform(
                        low=math.log(sys.float_info.min),
                        high=math.log(param.bound[1])))
                    config_setting[param.name] \
                        = np.random.choice([lb, ub])

                else:
                    # Bounds are positive
                    config_setting[param.name] \
                        = math.exp(np.random.uniform(
                            low=math.log(param.bound[0]),
                            high=math.log(param.bound[1])))
            else:
                config_setting[param.name] \
                    = np.random.uniform(low=param.bound[0],
                                        high=param.bound[1])

        elif param.type == ParamType.integer:
            if param.scale:
                # Generate in logarithmic space
                if param.bound[1] < 0:
                    # Upper and lower bound are negative
                    config_setting[param.name] \
                        = int(-1 * math.exp(np.random.uniform(
                            low=math.log(param.bound[0] * -1),
                            high=math.log(param.bound[1] * -1))))

                elif param.bound[0] < 0:
                    # Lower bound is negative
                    lb = int(-1 * math.exp(np.random.uniform(
                        low=math.log(sys.float_info.min),
                        high=math.log(param.bound[0] * -1))))
                    ub = int(math.exp(np.random.uniform(
                        low=math.log(sys.float_info.min),
                        high=math.log(param.bound[1]))))
                    config_setting[param.name] \
                        = np.random.choice([lb, ub])

                else:
                    # Bounds are positive
                    config_setting[param.name] \
                        = int(math.exp(np.random.uniform(
                            low=math.log(param.bound[0]),
                            high=math.log(param.bound[1]))))
            else:
                config_setting[param.name] \
                    = np.random.randint(low=param.bound[0],
                                        high=param.bound[1])

    return config_setting


def random_point(s, identity, seed=False):
    """
    Random parameter setting is generated in Configuration format.

    : param s: scenario object
    : param identity: uuid to identify configuration
    return: randomly set configuration, which accounts for no goods
    and conditionals
    """
    if seed:
        np.random.seed(seed)

    # Generate configuration randomly based on given parameter space
    config_setting = random_set_conf(s.parameter)

    # Check no goods and reset values if violated
    ng_vio = check_no_goods(s, config_setting)
    while ng_vio:
        config_setting = reset_no_goods(s, config_setting)
        ng_vio = check_no_goods(s, config_setting)

    # Fill Configuration class with ID and parameter values
    configuration = Configuration(identity, config_setting, Generator.random)

    return configuration



import ray
import logging
import time


@ray.remote(num_cpus=1)
class InstanceMonitor:
    def __init__(self, sleep_time, cache, scenario, delta_cap=1):
        """
        Monitor whether the runtime of a configuration on an instance exceeds the best runtime of any configuration
        on that instance multiplied by a constant (delta_cap).
        When a runtime exceeds this bound the configuration/instance pair is terminated.
        The terminated configuration/instance pairs are stored in the termination_history to avoid double killing
        :param sleep_time: Int. Wake up and check whether runtime is exceeded
        :param cache: Ray cache
        :param delta_cap: Int. Constant the current best runtime for each instance is multiplied by
        :return: conf/instance that are killed.
        """
        self.sleep_time = sleep_time
        self.cache = cache
        self.tournaments = []
        self.termination_history = {}
        self.best_instance_results = {}
        self.delta_cap = delta_cap
        self.last_t_check = 0
        self.scenario = scenario

        logging.basicConfig(filename=f'./selector/logs/{scenario.log_folder}/inst_monitor.log', level=logging.INFO,
                            format='%(asctime)s %(message)s')

    def monitor(self):
        logging.info("Starting monitor")
        worst_finisher = []

        while True:

            # Get results that are already available for ta runs
            start = time.time()
            results = ray.get(self.cache.get_results.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting results {dur}")

            # get starting times for each conf/instance
            start = time.time()
            start_time = ray.get(self.cache.get_start.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting start {dur}")

            # Get the current tournaments that are in the cache
            start = time.time()
            tournaments = ray.get(self.cache.get_tournament.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting tournaments {dur}")

            start = time.time()
            tournament_history = ray.get(self.cache.get_tournament_history.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting tournament history {dur}")

            current_instance_set = tournaments[0].instance_set_id
            tournaments_to_consider = len(tournaments)

            # finding the worst finisher for the tournaments with the most seen instances
            if len(tournament_history) != self.last_t_check and len(tournament_history) >= 2:
                last_tournaments = {}
                set_counter = {}
                for finished_tournament in tournament_history.values():
                    if finished_tournament.instance_set_id in  [current_instance_set, current_instance_set+1, current_instance_set-1]:
                        last_tournaments[finished_tournament.id] = finished_tournament
                        set_counter[finished_tournament.id] = finished_tournament.instance_set_id

                tournament_ids_to_consider = sorted(set_counter)[:tournaments_to_consider]
                worst_finisher = [last_tournaments[x].best_finisher[-1] for x in tournament_ids_to_consider]
                self.last_t_check = len(tournament_history)

            # Creating a dictionary containing the best runtime for each instance
            for conf in worst_finisher:
                # need to figure out which tournamnet has the highest instance number and then wich conf the smallest runtime
                for instance in results[conf.id]:
                    if instance in self.best_instance_results and \
                            results[conf.id][instance] < self.best_instance_results[instance]:
                        self.best_instance_results[instance] = results[conf.id][instance]
                    elif instance not in self.best_instance_results:
                        self.best_instance_results[instance] = results[conf.id][instance]

            logging.info(f"best_instance_results: {self.best_instance_results}")


            for t in tournaments:
                if len(t.ray_object_store.keys()) >= 1:
                    for conf in t.configurations:
                        instances_conf_finished = []
                        if conf.id in list(results.keys()):
                            instances_conf_finished = list(results[conf.id].keys())
                        instances_conf_planned = list(t.ray_object_store[conf.id].keys())
                        instances_conf_still_runs = [i for i in instances_conf_planned if i not in instances_conf_finished]

                        # We kill a configuration/instance pair, when the runtime exceeds the current best runtime so far
                        # for that instance multiplied by delta_cap or when the configuration timed out
                        for instance in instances_conf_still_runs:
                            if conf.id in start_time and instance in start_time[conf.id] \
                                    and instance in self.best_instance_results:
                                instance_runtime = time.time() - start_time[conf.id][instance]
                                logging.info(
                                    f"Monitor kill check: conf.id: {conf.id}, "
                                    f"instance: {instance}, "
                                    f"instance_runtime: {instance_runtime}, "
                                    f"best_instance_runtime: {self.best_instance_results[instance]}")
                                if instance_runtime > \
                                        self.best_instance_results[instance] * self.delta_cap:
                                    if self.termination_check(conf.id, instance):
                                        logging.info(
                                            f"Monitor is killing: {conf} {instance} "
                                            f"with id: {t.ray_object_store[conf.id][instance]}")
                                        print(f"Monitor is killing: {time.ctime()} {t.ray_object_store[conf.id][instance]}")
                                        self.update_termination_history(conf.id, instance)
                                        [ray.cancel(t.ray_object_store[conf.id][instance])]
                                    else:
                                        continue

            time.sleep(self.sleep_time)

    def termination_check(self, conf_id, instance):
        """
        Check if we have killed a conf/instance pair already. Return True if we did not.
        :param conf_id:
        :param instance:
        :return:
        """
        if conf_id not in self.termination_history:
            return True
        elif instance not in self.termination_history[conf_id]:
            return True
        else:
            return False

    def update_termination_history(self, conf_id, instance_id):
        if conf_id not in self.termination_history:
            self.termination_history[conf_id] = []

        if instance_id not in self.termination_history[conf_id]:
            self.termination_history[conf_id].append(instance_id)
        else:
            logging.info(f"This should not happen: we kill something we already killed")
